
/*

________________________________________________________________

        Image.c
        %W% %E%, Copyright 1990, Blab, UiO
        Image processing lab, Department of Informatics
        University of Oslo
        E-mail: blab@uio.no
________________________________________________________________
  
  Permission to use, copy, modify and distribute this software and its
  documentation for any purpose and without fee is hereby granted, 
  provided that this copyright notice appear in all copies and that 
  both that copyright notice and this permission notice appear in supporting
  documentation and that the name of B-lab, Department of Informatics or
  University of Oslo not be used in advertising or publicity pertaining 
  to distribution of the software without specific, written prior permission.

  B-LAB DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL B-LAB
  BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
  OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
  CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 

*/

static char *SccsId = "%Z%%M% %I% %E%, Otto Milvang, Blab, UiO";



/*D:Image*

________________________________________________________________

		ImageWidget
________________________________________________________________

Name:		Image Widget
Description:	| External Functions:
                |
                |   ImageTitle(img)
                |   struct imagestruct *img;
		|
		|   ImageAddress(band)
		|   struct bandstruct *band;
		|
		|   ImageWidth(band)
		|   struct bandstruct *band;
		|
		|   ImageHeight(band)
		|   struct bandstruct *band;
		|
		|   ImageLineLength(band)
		|   struct bandstruct *band;
		|
		|   ImageXstart(band)
		|   struct bandstruct *band;
		|
		|   ImageYstart(band)
		|   struct bandstruct *band;
		|
		|   ImagePixName(band)
		|   struct bandstruct *band;
		|
		| 
		| Public Functions:
                |
		|   void ImageError(wid, messageno)
		|   Widget wid;
		|   int messageno;
		|
		|   int ImageHasColormap(wid)
		|   ImageWidget wid;
		|
		|   int ImagePlanes(wid)
		|   ImageWidget wid;
		|
		|   int ImageGetDisplayMethod(defaultdm, class, planes, cells)
		|   int defaultdm, class, planes, cells;
		|
		|   void ImageInstallPixelType(pix)
		|   pixel pix;
		|
                |   void ImageRedisplay(wid, update)
		|   Widget wid;
		|   int update;
		| 
                |   void ImageRedisplayScreen(wid,  x, y, width, height)
		|   Widget wid;
		|   int x, y, width, height;
		| 
                |   void ImageRedisplayArea(wid, x, y, width, height)
		|   Widget wid;
		|   int x, y, width, height;
		| 
		|   void ImageKill(wid)
		|   Widget wid;
		|
		|   int ImageScreenXtoPos(wid, x)
		|   ImageWidget wid;
		|   int x;
		|
		|   int ImageScreenYtoPos(wid, y)
		|   ImageWidget wid;
		|   int y;
		|
		|   ImageData ImageScreenXYtoAdr(wid, x, y, option)
		|   ImageWidget wid;
		|   int x, y, option;
		|     
		|   ImageData ImagePosXYtoAdr(wid, x, y, option)
		|   ImageWidget wid;
		|   int x, y, option;
		|     
		|   void ImageTranslateZoomPanArgs(wid, zoomarg, panxarg,
		|        panyarg, mousex, mousey, zoom, xpan, ypan)
		|   ImageWidget wid;
		|   char *zoomarg, *panxarg, *panyarg;
		|   int  mousex, mousey,  *xpan, *ypan;
		|   float *zoom;
		|
		|   int ImagePosXtoScreen(wid, x, right)
		|   ImageWidget wid;
		|   int x, right;
		|
		|   int ImagePosYtoScreen(wid, y, bottom)
		|   ImageWidget wid;
		|   int y, bottom;
		|
		|   void ImageGetZoom(wid, zoom, xpan, ypan, factor)
		|   ImageWidget wid;
		|   float *zoom;
		|   int *xpan, *ypan;
		|   double *factor;
		|
		|   void ImageSetZoom(wid, zoom, x, y)
		|   Widget wid;
		|   double zoom;
		|   int x, y;
		|
		|   void ImageCopyImagePart(widto, widfrom)
		|   ImageWidget widto, widfrom;
		|
		|   int ImageGetZoomAll(wid, on)
		|   Widget wid;
		|   int on;
		|
		|   int ImageSetZoomAll(wid, on)
		|   Widget wid;
		|   int on;
		|
                | Private Functions:
		|
		|   static void ClassInitialize(wc)
		|   WidgetClass wc;
		|
                |   static void Initialize(request, new)
		|   Widget request, new;
		|
		|   static void Realize (wid, valueMask, attrs)
                |   Widget wid;
                |   XtValueMask *valueMask;
                |   XSetWindowAttributes *attrs;
		|
		|   static void Redisplay(wid, event, region)
		|   Widget wid;
		|   XEvent *event;
		|   Region region;
                |
		|   static XtGeometryResult Geometry(wid, request, answer)
		|   Widget wid;
		|   XtWidgetGeometry *request, *answer;
                |
		|   static void Nop(wid, event, params, num_params)
                |   Widget wid;
		|   XEvent *event;
		|   String *params;
		|   Cardinal *num_params;
		|
		|   static void Kill(wid, event, params, num_params)
                |   Widget wid;
		|   XEvent *event;
		|   String *params;
		|   Cardinal *num_params;
		|
		|   static void ZoomPan(wid, event, params, num_params)
                |   Widget wid;
		|   XEvent *event;
		|   String *params;
		|   Cardinal *num_params;
		|
		|   static void Button(wid, event, params, num_params)
                |   Widget wid;
		|   XEvent *event;
		|   String *params;
		|   Cardinal *num_params;
		|
		|   static void Resize(wid)
		|   Widget wid;
		|
		|   static Boolean SetValues(current, request, new)
		|   Widget current, request, new;
		|
		|   static Boolean SetValuesHook(wid, args, num_args)
		|   Widget wid;
		|   ArgList args;
		|   Cardinal *num_args;
		|
Description:
Files:	
See also:	
Diagnostics:
Restrictions:
Return value:
Author:		Otto Milvang
Examples:
________________________________________________________________

*/





/*I*

________________________________________________________________

		Include files and definitions
________________________________________________________________

*/


#include <X11/Xos.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include "ImageP.h"
#include <stdio.h>
#include <ctype.h>


#ifdef DEBUG 
#define FPRINTF(a,b,c,d) fprintf(a,b,c,d)
#else
#define FPRINTF(a,b,c,d)
#endif

#define Dist(a, b)    ( ((a) - (b)) < 0 ? ((b)-(a)) : ((a) - (b)) )
#define XimageWidth    ((ip->ximg)->width)
#define XimageHeight   ((ip->ximg)->height)
#define BandWidth      (ip->width)
#define BandHeight     (ip->height)

typedef struct 
{
  short x,y;
  unsigned short width, height;
  Window window
} EventRectangle;


char *malloc();

static void ClassInitialize();
static Boolean SetValues();
static Boolean SetValuesHook();
static void Initialize();
static void Realize();
static void Resize();
static void Redisplay();
static XtGeometryResult Geometry();
static void Button();
static void Kill();
static void Nop();
static void ZoomPan();

static int IfExpose();
static void PrintValues();
static void PrintUnknown();
static void SetZoomParam();
static double Resize_new_ww_wh();




/*I*

________________________________________________________________

		Predefined pixel types
________________________________________________________________

*/


int ImageByteResize();
char *ImageBytePrintf();
int ImageShortResize();
char *ImageShortPrintf();
int ImageUnsignedShortResize();
char *ImageUnsignedShortPrintf();
int ImageLongResize();
char *ImageLongPrintf();
int ImageFloatResize();
char *ImageFloatPrintf();
int ImageDoubleResize();
char *ImageDoublePrintf();


static pixel bytepixel =
{
  ImageBytePixel,
  1,
  ImageByteResize,
  ImageBytePrintf,
  3,
  SCALE_LINEAR,
  1.0,
  0.0,
  0.0,
  256.0,
};

static pixel shortpixel =
{
  ImageShortPixel,
  2,
  ImageShortResize,
  ImageShortPrintf,
  6,
  SCALE_LINEAR,
  1.0,
  0.0,
  -32768.0,
  32768.0,
};

static pixel unsignedshortpixel =
{
  ImageUnsignedShortPixel,
  2,
  ImageUnsignedShortResize,
  ImageUnsignedShortPrintf,
  5,
  SCALE_LINEAR,
  1.0,
  0.0,
  0.0,
  65536.0,
};


static pixel longpixel =
{
  ImageLongPixel,
  4,
  ImageLongResize,
  ImageLongPrintf,
  6,
  SCALE_LINEAR,
  1.0,
  0.0,
  -32768.0,
  32768.0,
};

static pixel floatpixel =
{
  ImageFloatPixel,
  4,
  ImageFloatResize,
  ImageFloatPrintf,
  6,
  SCALE_LINEAR,
  1.0,
  0.0,
  0.0,
  1.0,
};

static pixel doublepixel =
{
  ImageDoublePixel,
  8,
  ImageDoubleResize,
  ImageDoublePrintf,
  6,
  SCALE_LINEAR,
  1.0,
  0.0,
  0.0,
  1.0,
};




/*I*

________________________________________________________________

		Actions and tranlations
________________________________________________________________

*/



static XtActionsRec actions[] =
{
  /* {name, procedure}, */
  {"nop", 	Nop},
  {"notify",    Button},
  {"kill",	ImageKill},
  {"zoom_pan",	ZoomPan},
};

 static char translations[] =
"<Key>Q:             kill()\n\
 None<Key>Left:		zoom_pan(+0, -4, +0)\n\
 None<Key>Right:	zoom_pan(+0, +4, +0)\n\
 None<Key>Up:		zoom_pan(+0, +0, -4)\n\
 None<Key>Down:		zoom_pan(+0, +0, +4)\n\
 Shift<Key>Left:	zoom_pan(+0, -16, +0)\n\
 Shift<Key>Right:	zoom_pan(+0, +16, +)\n\
 Shift<Key>Up:		zoom_pan(+0, +0, -16)\n\
 Shift<Key>Down:	zoom_pan(+0, +0, +16)\n\
 Ctrl<Key>Left:		zoom_pan(+0, -256, +0)\n\
 Ctrl<Key>Right:	zoom_pan(+0, +256, +0)\n\
 Ctrl<Key>Up:		zoom_pan(+0, +0, -256)\n\
 Ctrl<Key>Down:		zoom_pan(+0, +0, +256)\n\
 Shift Ctrl <Btn1Down>:	zoom_pan(*8, MousePosC, MousePosC)\n\
 Shift Ctrl <Btn2Down>: zoom_pan(32, MousePosC, MousePosC)\n\
 Shift Ctrl <Btn3Down>: zoom_pan(/8, MousePosC, MousePosC)\n\
 Ctrl <Btn1Down>:       zoom_pan(*2, MousePosC, MousePosC)\n\
 Ctrl <Btn2Down>:       zoom_pan(1, MousePosC, MousePosC)\n\
 Ctrl <Btn3Down>:       zoom_pan(/2, MousePosC, MousePosC)\n\
 None <Btn1Up>:         notify()\n\
 None <Btn2Up>:         notify()\n\
 None <Btn3Up>:         notify()\n\
 <MouseMoved>:          notify()";




/*I*

________________________________________________________________

		Resources
________________________________________________________________

*/



static XtResource resources[] = {
#define ioffset(field) XtOffset(ImageWidget, image.field)
#define coffset(field) XtOffset(ImageWidget, core.field)
    /* {name, class, type, size, ioffset, default_type, default_addr}, */
    { XtNimageImage, XtCImageImage, XtRPointer, sizeof(char *),
	  ioffset(image), XtRPointer , NULL }, 
    { XtNimageBand, XtCImageBand, XtRPointer, sizeof(char *),
	  ioffset(band), XtRPointer, NULL }, 
    { XtNimageRedBand, XtCImageRedBand, XtRPointer, sizeof(char *),
	  ioffset(red_band), XtRPointer, NULL }, 
    { XtNimageGreenBand, XtCImageGreenBand, XtRPointer, sizeof(char *),
	  ioffset(green_band), XtRPointer, NULL }, 
    { XtNimageBlueBand, XtCImageBlueBand, XtRPointer, sizeof(char *),
	  ioffset(blue_band), XtRPointer, NULL }, 
    { XtNimageData, XtCImageData, XtRPointer, sizeof(char *),
	  ioffset(data), XtRPointer, NULL },
    { XtNimageRedData, XtCImageRedData, XtRPointer, sizeof(char *),
	  ioffset(red_data), XtRPointer, NULL },
    { XtNimageGreenData, XtCImageGreenData, XtRPointer, sizeof(char *),
	  ioffset(green_data), XtRPointer, NULL },
    { XtNimageBlueData, XtCImageBlueData, XtRPointer, sizeof(char *),
	  ioffset(blue_data), XtRPointer, NULL },
    { XtNimageTitle, XtCImageTitle, XtRPointer, sizeof(char *),
	  ioffset(title), XtRString, NULL }, 
    { XtNimagePixtype, XtCImagePixtype, XtRPointer, sizeof(char *),
	  ioffset(pixname), XtRString, NULL }, 
    { XtNimageWidth, XtCImageWidth, XtRInt, sizeof(int),
	  ioffset(width), XtRInt, NULL },
    { XtNimageHeight, XtCImageHeight, XtRInt, sizeof(int),
	  ioffset(height), XtRInt, NULL }, 
    { XtNimageLineLength, XtCImageLineLength, XtRInt, sizeof(int),
	  ioffset(line_length), XtRInt, NULL }, 
    { XtNimageXstart, XtCImageXstart, XtRInt, sizeof(int),
	  ioffset(xstart), XtRInt, NULL }, 
    { XtNimageYstart, XtCImageYstart, XtRInt, sizeof(int),
	  ioffset(ystart), XtRInt, NULL }, 
    { XtNdisplayMethod, XtCDisplayMethod, XtRInt, sizeof(int),
	  ioffset(displaymethod), XtRInt, NULL },
    { XtNprintValues, XtCPrintValues, XtRBoolean, sizeof(int),
	  ioffset(print_values), XtRImmediate, (XtPointer) TRUE },
    { XtNcolortab, XtCColortab, XtRInt, sizeof(int),
	  ioffset(colortab), XtRInt, NULL },
    { XtNvisual, XtCVisual, XtRVisual, sizeof(Visual*),
	  ioffset(visual), XtRPointer, NULL},
    { XtNscaletype, XtCScaletype, XtRInt, sizeof(int),
	  ioffset(scaletype), XtRInt, NULL },
    { XtNscalea, XtCScalea, XtRFloat, sizeof(float),
	  ioffset(scalea), XtRFloat, NULL },
    { XtNscaleb, XtCScaleb, XtRFloat, sizeof(float),
	  ioffset(scaleb), XtRFloat, NULL },
    { XtNminval, XtCMinval, XtRFloat, sizeof(float),
	  ioffset(minval), XtRFloat, NULL },
    { XtNmaxval, XtCMaxval, XtRFloat, sizeof(float),
	  ioffset(maxval), XtRFloat, NULL },
    {XtNfont,  XtCFont, XtRFontStruct, sizeof(XFontStruct *),
          ioffset(font[0]),XtRString, "XtDefaultFont"},
    {XtNfont1,  XtCFont1, XtRFontStruct, sizeof(XFontStruct *),
          ioffset(font[1]),XtRString, NULL},
    {XtNfont2,  XtCFont2, XtRFontStruct, sizeof(XFontStruct *),
          ioffset(font[2]),XtRString, NULL},
    {XtNfont3,  XtCFont3, XtRFontStruct, sizeof(XFontStruct *),
          ioffset(font[3]),XtRString, NULL},
    { XtNcallback, XtCCallback, XtRCallback, sizeof(caddr_t),
	  ioffset(button_action), XtRCallback, (caddr_t) NULL },
    { XtNresizeCallback, XtCResizeCallback, XtRCallback, sizeof(caddr_t),
	  ioffset(resize_action), XtRCallback, (caddr_t) NULL },

#undef ioffset
#undef coffset
};

#define SuperClass ((ImageWidgetClass)&imageClassRec)




/*I*

________________________________________________________________

		Image Class Record
________________________________________________________________

*/


ImageClassRec imageClassRec = {
  {
    /* core fields */
    /* superclass		*/	(WidgetClass) &compositeClassRec,
    /* class_name		*/	"Image",
    /* widget_size		*/	sizeof(ImageRec),
    /* class_initialize		*/	ClassInitialize,
    /* class_part_initialize	*/	NULL,
    /* class_inited		*/	FALSE,
    /* initialize		*/	Initialize,
    /* initialize_hook		*/	NULL,
    /* realize			*/	Realize,
    /* actions			*/	actions,
    /* num_actions		*/	XtNumber(actions),
    /* resources		*/	resources,
    /* num_resources		*/	XtNumber(resources),
    /* xrm_class		*/	NULLQUARK,
    /* compress_motion		*/	TRUE,
    /* compress_exposure	*/	TRUE,
    /* compress_enterleave	*/	TRUE,
    /* visible_interest		*/	FALSE,
    /* destroy			*/	Kill,
    /* resize			*/	Resize,
    /* expose			*/	Redisplay,
    /* set_values		*/	SetValues,
    /* set_values_hook		*/	SetValuesHook,
    /* set_values_almost	*/	XtInheritSetValuesAlmost,
    /* get_values_hook		*/	NULL,
    /* accept_focus		*/	NULL,
    /* version			*/	XtVersion,
    /* callback_private		*/	NULL,
    /* tm_table			*/	translations,
    /* query_geometry		*/	Geometry,
    /* display_accelerator	*/	XtInheritDisplayAccelerator,
    /* extension		*/	NULL
  },{
    /* composite_class fields */
    /* geometry_manager   	*/    	XtInheritGeometryManager,
    /* change_managed     	*/    	XtInheritChangeManaged,
    /* insert_child       	*/    	XtInheritInsertChild,
    /* delete_child       	*/    	XtInheritDeleteChild,
    /* extension          	*/    	NULL,
  },{
    /* image fields */
    /* count			*/	0,
    /* widget_list_length       */      0,
    /* widget_list              */      NULL,
    /* pixel class              */	NULL,
    /* zoom_all                 */      0,
  }
};

WidgetClass imageWidgetClass = (WidgetClass)&imageClassRec;




/*L*

________________________________________________________________

		ImageError
________________________________________________________________

Name:		ImageError
Syntax:		
Description:
Return value:
Author:		Otto Milvang
________________________________________________________________

*/

#define ERR_OK		0
#define ERR_MEM		1
#define ERR_PARAM	2
#define ERR_DM		3
#define ERR_PT		4

static String err_mess[] = {
  "ImageWidget: No error",
  "ImageWidget: Not enough memory",
  "ImageWidget: Bad parameters to zoom_pan",
  "ImageWidget: Unknown DisplayMethod",
  "ImageWidget: Unknown Pixeltype",
};

void ImageError(wid, messnr)
Widget wid;
int messnr;
{
  XtAppErrorMsg(
    XtWidgetToApplicationContext(wid),
    "Error","ImageWidget","XtToolkitError",
    err_mess[messnr],
    (String *)NULL, (Cardinal *)NULL);
}





/*L:ImageHasColormap*

________________________________________________________________

		ImageHasColormap
________________________________________________________________

Name:		ImageHasColormap
Syntax:		| ImageHasColormap(wid)
                | ImageWidget wid;
Description:	
Return value:	| True if Visual Type is StaticGray, StaticColor or TrueColor
                | False if Visual type is GrayScale, PseudoColor or DirectColor
Author:		Otto Milvang
________________________________________________________________

*/



ImageHasColormap(wid)
ImageWidget wid;
{
  return(wid->image.has_colormap);
}





/*L:ImagePlanes*

________________________________________________________________

		ImagePlanes
________________________________________________________________

Name:		ImagePlanes
Syntax:		| ImagePlanes(wid)
                | ImageWidget wid;
Description:
Return value:	| Number of planes 
Author:		Otto Milvang
________________________________________________________________

*/


ImagePlanes(wid)
ImageWidget wid;
{
  return(wid->image.planes);
}






/*L:ImageGetDisplayMethod*

________________________________________________________________

		ImageGetDisplayMethod
________________________________________________________________

Name:		ImageGetDisplayMethod
Syntax:		| ImageGetDisplayMethod(defaultdm, class, planes, cells)
                | int defaultdm, class, planes, cells;
Description:	Find the display method from class, planes and cells;
Return value:	New display method.

Author:		Otto Milvang
________________________________________________________________

*/


ImageGetDisplayMethod(defaultdm, class, planes, cells)
int defaultdm, class, planes, cells;
{
  int dm;
  dm = defaultdm < 16 ? class*0x1000 + planes * 0x10 :  defaultdm;
  if (defaultdm < 16 && (class == PseudoColor || class == GrayScale ||
			class == DirectColor)) dm = dm + defaultdm;
  FPRINTF(stderr, "Test dm = %d\n", dm, 0);
  if (dm/16 != defaultdm/16 && defaultdm >=16) return(0);
  if (dm == STATIC_GRAY_1) return(STATIC_GRAY_1);
  if (dm == STATIC_GRAY_2) return(STATIC_GRAY_2);
  if (dm == STATIC_GRAY_4) return(STATIC_GRAY_4);
  if (dm == STATIC_GRAY_8) return(STATIC_GRAY_8);
  if (dm == GRAY_SCALE_4) return(GRAY_SCALE_4);
  if (dm == GRAY_SCALE_8) return(GRAY_SCALE_8);
  if (dm == PSEUDO_COLOR_8_A) return(PSEUDO_COLOR_8_A);
  if (dm == PSEUDO_COLOR_8_B) return(PSEUDO_COLOR_8_B);
  if (dm == TRUE_COLOR_24) return(TRUE_COLOR_24);

  return(0);
}





/*L:ImageInstallPixelType*

________________________________________________________________

		ImageInstallPixelType
________________________________________________________________

Name:		ImageInstallPixelType
Syntax:		| void ImageInstallPixelType(pix)
                | pixel *pix;
Description:
Return value:	 
Author:		Otto Milvang

*/

void ImageInstallPixelType(pix)
pixel *pix;
{
  ImageClassPart *ic = &imageClassRec.image_class;
  pixelptr ptr;
  ptr = (pixelptr) malloc(sizeof(ipixel));
  if (ptr == NULL) XtError("malloc() failed");
  ptr->pixelname = pix->pixelname;
  ptr->pixelsize = pix->pixelsize;
  ptr->resize = pix->resize;
  ptr->printf = pix->printf;
  ptr->txtlen = pix->txtlen;
  ptr->scaletype = pix->scaletype;
  ptr->scalea = pix->scalea;
  ptr->scaleb = pix->scaleb;
  ptr->minval = pix->minval;
  ptr->maxval = pix->maxval;

  ptr->next = ic->pixel_class;
  ic->pixel_class = ptr;
}




/*L*

________________________________________________________________

		ClassInitialize
________________________________________________________________

Name:		ClassInitialize
Syntax:		| static void ClassInitialize(wc)
                | WidgetClass wc;
Description:    
Return value:
Author:		Otto Milvang
________________________________________________________________

*/


static void ClassInitialize(wc)
WidgetClass wc;
{
  ImageClassPart *ic = &imageClassRec.image_class;
  XawInitializeWidgetSet();
  ic->widget_list_length = 16;
  ic->widget_list = (Widget *) malloc(16 * sizeof(Widget));
  if (ic->widget_list == NULL) XtError("malloc() failed");
  FPRINTF(stderr, "Image: Class Initialize %s\n", "Image",0);
  ImageInstallPixelType(&bytepixel);
  ImageInstallPixelType(&shortpixel);
  ImageInstallPixelType(&unsignedshortpixel);
  ImageInstallPixelType(&longpixel);
  ImageInstallPixelType(&floatpixel);
  ImageInstallPixelType(&doublepixel);
} 




/*L*

________________________________________________________________

		Initialize
________________________________________________________________

Name:		Initialize
Syntax:		| static void Initialize(request, new)
                | Widget request, new;
		|
Description:
Return value:
Author:		Otto Milvang
________________________________________________________________

*/


static void Initialize(request, new)
Widget request, new;
{
  ImageWidget iw = (ImageWidget) new;
  ImageClassPart *ic = &imageClassRec.image_class;
  ImagePart *ip;
  int i, j, w, h, c, xpan, ypan;
  Display *dpy;
  float zoom;
  int screen, size;
  ImageWidget par;
  unsigned char *pt1, *pt2;
  XFontStruct        *fh;
  pixelptr pptr;

  dpy = XtDisplay(new);
  screen = DefaultScreen(dpy);

  if (new == NULL) return;
  
  FPRINTF(stderr, "Image: Initialize (%d %d) \n",iw->image.xstart,iw->image.ystart);
  if (new == NULL) return;
  ip = &(iw->image);

  if (ip->band == NULL && ip->green_band != NULL) ip->band = ip->green_band;
  if (  (ip->data == 0   && ip->band == 0)
     || (ip->width == 0  && ip->band == 0)
     || (ip->height == 0 && ip->band == 0) )
    XtErrorMsg("Image", "Initialize", "Image",
    "Unknown image address, width or height", NULL,0);

  if (ip->title == 0)
    {
      ip-> title = malloc(33);
      if (ip->title == NULL) ImageError(new, ERR_MEM);
      ip-> title[0] = 0;
      if (ip->image) strcpy(ip->title, ImageTitle(ip->image));
      ip->title[32] = 0;
    }
  i = strlen(ip->title);
  while (i > 0 && ip->title[i] == ' ') ip->title[i--] = 0;
  if (strlen(ip->title) == 0) strcpy(ip->title,"Image");

  if (ip->band)
  {
     ip->data        = ImageAddress(ip->band);
     ip->pixname     = ImagePixName(ip->band);
     ip->width       = ImageWidth(ip->band);
     ip->height      = ImageHeight(ip->band);
     ip->line_length = ImageLineLength(ip->band);
     ip->xstart      = ImageXstart(ip->band);
     ip->ystart      = ImageYstart(ip->band);
  }
  if (ip->red_band)	ip->red_data = ImageAddress(ip->red_band);
  if (ip->green_band)	ip->green_data = ImageAddress(ip->green_band);
  if (ip->blue_band)	ip->blue_data = ImageAddress(ip->blue_band);

  if (ip->data == NULL && ip->green_data != NULL) ip->data = ip->green_data;
  if (ip->pixname)
    {
      pptr = ic->pixel_class;
      while (pptr && strcmp(pptr->pixelname, ip->pixname)) pptr = pptr->next;
    } else pptr = NULL;
  ip->pixtype = pptr;
  /*  if (pptr == NULL) ImageError(iw, ERR_PT); */

  if (ip->line_length == 0) ip->line_length = ip->width * 
    (pptr ? pptr->pixelsize : 1);
 
 /* while (wid->core.parent) wid = wid->core.parent; */
  if (ip->visual == NULL)
    {
      FPRINTF(stderr,"Use default visual !!!!!!!!!!!!!!!\n",0,0);
      ip->visual = DefaultVisual(dpy,screen);
    }
  c = ip->visual->class;
  ip->has_colormap = (c == PseudoColor || c == GrayScale || c == DirectColor);
  ip->planes = iw->core.depth;
  ip->cells = ip->visual->map_entries;
  ip->displaymethod = 
    ImageGetDisplayMethod(ip->displaymethod, c, ip->planes, ip->cells);
  FPRINTF(stderr, "Displaymethod : ...............%d\n", ip->displaymethod, 0);
      FPRINTF(stderr, "Visual %d %d", ip->has_colormap, ip->cells);
  FPRINTF(stderr, " %d \n", ip->planes, 0);
  if (!ip->displaymethod) ImageError(iw, ERR_DM);
  w = ip -> width;
  h = ip -> height;
  ip -> factor = 1.0;
  ip -> resize = FALSE;
  ip -> update = FALSE;
  ip->zoom = 1.0;
  ip->xpan = 0;
  ip->ypan = 0;
  if (pptr && ip->scalea == 0.0 && ip->scaleb == 0.0)
    {
      ip->scalea = pptr->scalea;
      ip->scaleb = pptr->scaleb;
    }
  if (pptr && ip->minval == 0.0 && ip->maxval == 0.0)
    {
      ip->minval = pptr->minval;
      ip->maxval = pptr->maxval;
    }
  ip -> ximg = XCreateImage(dpy,
               ip->visual,
               iw->core.depth,
               ip -> displaymethod == STATIC_GRAY_1 ? XYBitmap : ZPixmap , 
               0, NULL, 0, 0, 8, 0);

  (ip -> ximg)->bitmap_bit_order = BitmapBitOrder(dpy);
  par = (ImageWidget) iw->core.parent;
  ip -> xptr = NULL;
  ip -> yptr = NULL;
  ip->internal_width = iw->core.width;
  ip->internal_height = iw->core.height;
  while (w > 1000 || h > 800) 
    {
      w >>= 1;
      h >>= 1;
    }
  if (ic->zoom_all && ic->count)
    {
      ImageGetZoom(ic->widget_list[0], &zoom, &xpan, &ypan, NULL);
      SetZoomParam(iw, zoom, xpan, ypan);
    }
  if (strcmp((iw->core.widget_class)->core_class.class_name, "Image") == 0)
    {
      if (iw->core.width == 0 || iw->core.height == 0)
       XtResizeWidget(new, w, h, 1); else
       Resize(new);
    }
  if (ic->count>= ic->widget_list_length)
    {
      ic->widget_list_length += 16;
      ic->widget_list = 
	(Widget *) realloc(ic->widget_list, 
			   ic->widget_list_length * sizeof(Widget));
    }
  ip->gc = DefaultGC(dpy, screen); /* In case Widget is'nt realized */
  for(i=1;  i<=3; i++) if (ip->font[i] == NULL) ip->font[i] = ip->font[0];
  for (i=0; i<=3; i++) FPRINTF(stderr,"Font %d -> %d\n",i,ip->font[i]);
  for (i=1; i<=2; i++) for (j=i+1; j<=3; j++)
    if (ip->font[i]->max_bounds.width > ip->font[j]->max_bounds.width)
      {
	fh = ip->font[i];
	ip->font[i] = ip->font[j];
	ip->font[j] = fh;
      }
  ic->widget_list[ic->count] = new;
  ic->count ++;
  FPRINTF(stderr, "Image: Initialize end\n",0,0);
}

 




/*L*

________________________________________________________________

		Realize
________________________________________________________________

Name:		Realize
Syntax:		| static void Realize (wid, valueMask, attrs)
                | Widget wid;
                | XtValueMask *valueMask;
                | XSetWindowAttributes *attrs;
Description:
Return value:
Author:		Otto Milvang
________________________________________________________________

*/

static void Realize (wid, valueMask, attrs)
     ImageWidget wid;
     XtValueMask *valueMask;
     XSetWindowAttributes *attrs;
{
  Display *dpy;
  int screen;
  Colormap map;
  ImageWidget        iw = (ImageWidget) wid;
  XGCValues gcvalues;

  FPRINTF(stderr, "Image: Realize %04x \n",*valueMask, 0);
  dpy = XtDisplay(wid);
  screen = DefaultScreen(dpy);
  gcvalues.foreground = WhitePixel(dpy, screen);
  gcvalues.background = BlackPixel(dpy, screen);
  XtCreateWindow(wid, InputOutput, wid->image.visual, *valueMask , attrs);
  wid->image.gc = 
    XCreateGC(dpy, XtWindow(wid),
	      GCForeground | GCBackground, &gcvalues);
  XDefineCursor(dpy, XtWindow(wid), XCreateFontCursor(dpy,XC_crosshair));
}



/*L*

________________________________________________________________

		Kill
________________________________________________________________

Name:		Kill
Syntax:		| static void Kill(wid, event, params, num_params)
                | Widget wid;
		| XEvent *event;
		| String *params;
		| Cardinal *num_params;
		
Description:
Return value:
Author:		Otto Milvang
________________________________________________________________

*/



static void Kill(wid,event,params,num_params)
ImageWidget wid;
XEvent *event;
String *params;
Cardinal *num_params;
{
  ImagePart *ip;
  int i;
  ImageClassPart *ic = &imageClassRec.image_class;
  ip = &(wid->image);

  FPRINTF(stderr, "Image: Kill\n",0,0);
  i = 0;
  ic->count--;
  while(i < ic->count && ic->widget_list[i] != (Widget) wid) i++;
  ic->widget_list[i] = ic->widget_list[ic->count];  
  if (ip->xptr) free(ip->xptr);
  if (ip->yptr) free(ip->yptr); 
  XDestroyImage(ip->ximg);
}



/*L*

________________________________________________________________

		Resize
________________________________________________________________

Name:		Resize
Syntax:		| static void Resize(wid)
                | Widget wid;
Description:
Return value:
Author:		Otto Milvang
________________________________________________________________

*/



static void Resize(wid)
ImageWidget wid;
{
  ImageWidget iw = (ImageWidget) wid;
  ImagePart *ip;
  struct bandstruct  *b;
  int h, w, i, size_changed;
  XtGeometryResult stat;
  double factor;
  int *xarr, *yarr, x, y, rw, rh;
  Widget child;

  ip = &(iw->image);
  b = ip->band;
  w = iw->core.width;
  h = iw->core.height;
  FPRINTF(stderr, "Image: Resize band %d \n", ip->band, 0);

  if (ip->update != 2)
    {
      factor = Resize_new_ww_wh(wid, &w, &h, 
				XimageWidth, XimageHeight,
				BandWidth, BandHeight, ip->update, ip->zoom);
      if (factor == 0.0) return;

      
      FPRINTF(stderr, "Image: Resize (%d, %d) ", iw->core.width ,iw->core.height );
      FPRINTF(stderr, " (%d, %d) \n",  w, h);

      size_changed = 0;

      if ((ip->ximg)->data != NULL) free((ip->ximg)->data);
      (ip->ximg)->data = NULL;
      if (ip->xptr) free(ip->xptr);
      if (ip->yptr) free(ip->yptr); 

      ip->xptr = xarr = (int *) malloc(w * sizeof(int));
      ip->yptr = yarr = (int *) malloc(h * sizeof(int));
      if (xarr == NULL || yarr == NULL) ImageError(wid, ERR_MEM);
      for(x=0; x < w; x++)  
	{
	  xarr[x] = (0.5 + x) / factor + ip->xpan;
	  if (xarr[x] < 0) xarr[x] = 0;
	  if (xarr[x] >= BandWidth) xarr[x] = BandWidth-1;
	}
      for(y=0; y < h; y++)
	{ 
	  yarr[y] = (0.5 + y) / factor + ip->ypan;
	  if (yarr[y] < 0) yarr[y] = 0;
	  if (yarr[y] >= BandHeight) yarr[y] = BandHeight-1;
	}

      
      if (w == BandWidth && h == BandHeight && ip->zoom == 1.0) 
	ip -> resize = FALSE; else
	ip -> resize = TRUE;
     
      (ip->ximg) -> width = w;
      (ip->ximg) -> height = h;
      if (ip->pixtype) (ip->pixtype)->
	resize(wid, 0, 0, wid->core.width, wid->core.height);
      ip -> factor = factor;
      if (ip->internal_width != w || ip->internal_height != h)
	{
	  stat = XtMakeResizeRequest(wid, w, h, &rw, &rh); 
	  ip->update = TRUE;
	  ip->internal_width = w;
	  ip->internal_height = h;
	  size_changed = 1;
	  for(i=0; i< wid->composite.num_children; i++)
	    XtResizeWidget(wid->composite.children[i],
			   iw->core.width, iw->core.height, 0);
	} /* XtResizeWidget(wid->core.parent, w, h, 1); */
      Button(wid, NULL, NULL, 0);
    } else {
      if (ip->pixtype) (ip->pixtype)->
	resize(wid, ip->updatex, ip->updatey, ip->updatew, ip->updateh);
    }
  ip->update = FALSE;
}





/*L:PrintValues*

________________________________________________________________

		PrintValues
________________________________________________________________

Name:		PrintValues
Syntax:		
Description:
Return value:
Author:		Otto Milvang
________________________________________________________________

*/


static void PrintValues(wid, x, y, width, height)
ImageWidget wid;
int x, y, width, height;
{
  Display *dpy;
  int screen;
  char val[8];
  int x1, x2, y1, y2;
  int xw, xh;
  int *xarr, *yarr;
  int *xs, *ys, cx, cy, xb, xl, yb, yl;
  int  fontn, fontw, fonth, fonta, bg, fg;
  ImagePart *ip;
  ip = &(wid->image);
  for (fontn = 3; fontn>=0; fontn--)
    {
      fontw = (ip->font[fontn])->max_bounds.width;
      fonta = (ip->font[fontn])->ascent;
      fonth = fonta + (ip->font[fontn])->descent;
      if (ip->factor >= ((0.2 + (ip->pixtype)->txtlen) * fontw)) break;
    }
  if (fontn <= 0) return; 

  dpy = XtDisplay(wid);
  screen = DefaultScreen(dpy);
  FPRINTF(stderr, "VFont:%d\n", fontn, 0);
  XSetFont(dpy, ip->gc, (ip->font[fontn])->fid);

  x1 = x;
  x2 = x + width - 1;
  y1 = y;
  y2 = y + height - 1;
  xw = (ip->ximg)->width-1;
  xh = (ip->ximg)->height-1;
  xarr = ip->xptr;
  yarr = ip->yptr;

  while (x1 >  0 && xarr[x1] == xarr[x1-1]) x1--; 
  while (y1 >  0 && yarr[y1] == yarr[y1-1]) y1--; 
  while (x2 < xw && xarr[x2] == xarr[x2+1]) x2++; 
  while (y2 < xh && yarr[y2] == yarr[y2+1]) y2++;
 

  xs = (int *) malloc((xarr[x2]-xarr[x1]+2) * sizeof(int));
  ys = (int *) malloc((yarr[y2]-yarr[y1]+2) * sizeof(int));
  if (xs == NULL || ys == NULL) ImageError(wid, ERR_MEM);
  cx = cy = 0;
  xs[cx++] = x1;
  for (x=x1+1; x <= x2; x++)
    if (ip->xptr[x] != ip->xptr[x-1]) xs[cx++] = x;
  xs[cx] = x2+1;
  ys[cy++] = y1;
  for (y=y1+1; y <= y2; y++)
    if (ip->yptr[y] != ip->yptr[y-1]) ys[cy++] = y;
  ys[cy] = y2+1;

  for(y=0; y<cy; y++)
    for(x=0; x<cx; x++)
      {            
	xb = xs[x];
	xl = xs[x+1] - xs[x];
	yb = ys[y];
	yl = ys[y+1] - ys[y];
	if (xl > fontw*2 && yl > fonth*2/3)
	  {
	    (ip->pixtype)->printf(val, ImageScreenXYtoAdr(wid, xb, yb, 0)); 
	    bg = XGetPixel(ip->ximg, xb, yb);
	    switch(ip->displaymethod)
	      {
	      case STATIC_GRAY_1:
		bg = bg ? BlackPixel(dpy,screen):WhitePixel(dpy,screen);
		fg = bg ? WhitePixel(dpy,screen):BlackPixel(dpy,screen);
		break;
	      case STATIC_GRAY_4:
		fg = bg < 8 ? bg + 8 : bg - 8; 
		break;
	      case PSEUDO_COLOR_8_A: 
		fg = bg < 128 ? bg + 64 : bg - 64; 
		break;
	      case STATIC_GRAY_8:
	      case PSEUDO_COLOR_8_B: 
		fg = bg < 128 ? bg + 128 : bg - 128;
		break;
	      }
	    XSetState (dpy,
		       ip->gc, 
		       fg, 
		       bg, 
		       GXcopy, 
		       AllPlanes); 
	    XDrawString(dpy, XtWindow(wid), ip->gc, 
			xb + (xl - strlen(val)*fontw)/2,
			yb + (yl - fonth)/2 + fonta, val, strlen(val));
	  } 
      }
  free(xs);
  free(ys);
}





/*L:PrintUnknown*

________________________________________________________________

		PrintUnknown
________________________________________________________________

Name:		PrintUnknown
Syntax:		
Description:
Return value:
Author:		Otto Milvang
________________________________________________________________

*/


static void PrintUnknown(wid)
ImageWidget wid;
{
  Display *dpy;
  int screen;
  int  fontn, fontw, fonth, fonta, bg, fg;
  ImagePart *ip;
  char txt[100];
  XColor exact, best;

  ip = &(wid->image);
  fontn = 2;
  fontw = (ip->font[fontn])->max_bounds.width;
  fonta = (ip->font[fontn])->ascent;
  fonth = fonta + (ip->font[fontn])->descent;

  dpy = XtDisplay(wid);
  screen = DefaultScreen(dpy);
  FPRINTF(stderr, "VFont:%d\n", fontn, 0);
  XSetFont(dpy, ip->gc, (ip->font[fontn])->fid);

  if (ip->pixname)
    sprintf(txt," %s - image ", ImagePixName(ip->band)); else
      sprintf(txt," Unknown pixtype ");
  XAllocNamedColor(dpy, wid->core.colormap, "White", &exact, &best);
  bg = best.pixel;
  XAllocNamedColor(dpy, wid->core.colormap, "Black", &exact, &best);
  fg = best.pixel; 
  FPRINTF(stderr, "Print text %s\n", txt, 0);
  XSetState (dpy,
	     ip->gc, 
	     fg, 
	     bg, 
	     GXcopy, 
	     AllPlanes);
  XDrawString(dpy,XtWindow(wid), ip->gc, 
	      (wid->core.width  - strlen(txt)*fontw)/2,
	      (wid->core.height - fonth)/2 + fonta, txt, strlen(txt));
}





/*L*

________________________________________________________________

		Redisplay 
________________________________________________________________

Name:		Redisplay
Syntax:		| static void Redisplay(wid, event, region)
                | Widget wid;
		| XEvent *event;
		| Region region;
		|
Description:    
Return value:
Author:		Otto Milvang
________________________________________________________________

*/


static int IfExpose(display, event, arg)
Display *display;
XEvent *event;
EventRectangle *arg;
{
  XExposeEvent *xevent;
  short start, stop;
  if (event->type != Expose) return(0);
  xevent = &event->xexpose;
  if (xevent->window != arg->window) return(0);
  start = arg->x < xevent->x ? arg->x : xevent->x;
  stop  = (arg->x + arg->width) > (xevent->x + xevent->width) ? 
          (arg->x + arg->width) : (xevent->x + xevent->width);
  arg->x = start;
  arg->width = stop - start;
  start = arg->y < xevent->y ? arg->y : xevent->y;
  stop  = (arg->y + arg->height) > (xevent->y + xevent->height) ? 
          (arg->y + arg->height) : (xevent->y + xevent->height);
  arg->y = start;
  arg->height = stop - start;
  return(1);
}  


static void Redisplay(wid, event, region)
Widget wid;
XEvent *event;
Region region;
{
  XEvent hevent;
  EventRectangle rect;
  ImageWidget iw = (ImageWidget) wid;
  ImagePart *ip;
  int *xs, *ys, maxx, maxy, x, y, cx, cy, xb, xl, yb, yl;
  int  fontn, fontw, fonth, fonta, bg, fg;
  char val[8], txt[100];
  Display *dpy;
  int screen, display;
  ImageWidget par;

    FPRINTF(stderr, "Image: Redisplay\n", 0,0);
  if (!XtIsRealized(wid)) return;
  dpy = XtDisplay(wid);
  screen = DefaultScreen(dpy);
  ip = &(iw->image);
  if (region) XClipBox(region, &rect); else
    {
       rect.x = 0;
       rect.y = 0;
       rect.width = (ip->ximg)->width;
       rect.height = (ip->ximg)->height;
     }
  rect.window = XtWindow(wid);

    FPRINTF(stderr, "Image: Redisplay x = %d %d, ",rect.x, rect.x+rect.width-1);
    FPRINTF(stderr, "    y = %d %d ",rect.y,rect.y+rect.height-1);
    FPRINTF(stderr, " ----- Address %08x\n", wid, 0);
  while (XCheckIfEvent(dpy, &hevent, IfExpose, &rect));
  if (rect.width  > (ip->ximg)->width)  rect.width  = (ip->ximg)->width;
  if (rect.height > (ip->ximg)->height) rect.height = (ip->ximg)->height;
  par = (ImageWidget) iw->core.parent;
  if (ip->pixtype)
    {
      XSetState (dpy,
		 ip->gc, 
		 /*ip->foreground */ WhitePixel(dpy,screen), 
		 /*ip->background */ BlackPixel(dpy,screen), 
		 GXcopy, 
		 AllPlanes); 
 
      XPutImage (dpy, 
             XtWindow(wid),
             ip->gc,
             ip->ximg,
             rect.x,
             rect.y,
             rect.x,
             rect.y,
             rect.width,
             rect.height);
      if (ip->print_values)
	PrintValues(wid, rect.x, rect.y, rect.width, rect.height);
    } else PrintUnknown(wid);
}  








/*L*

________________________________________________________________

		SetValues
________________________________________________________________

Name:		SetValues
Syntax:		
Description:
Return value:
Author:		Otto Milvang
________________________________________________________________

*/


static Boolean SetValues(current, request, new)
Widget current, request, new;
{
  pixelptr pptr;
  ImageWidget ic = (ImageWidget) current;
  ImageWidget ir = (ImageWidget) request;
  ImageWidget in = (ImageWidget) new;
  ImageClassPart *iic = &imageClassRec.image_class;
  
  int i, w, h;
  Display *dpy;
  int screen;

  dpy = XtDisplay(new);
  screen = DefaultScreen(dpy);

  if (ic->image.visual != ir->image.visual)
    XtErrorMsg("Image", "Initialize", "Image",
    "Can't change Visual", NULL,0);

  FPRINTF(stderr, "Image: Set Values %10.2f %10.2f \n",in->image.minval,in->image.maxval);
  if (ic->image.band != ir->image.band)
  {
     in->image.data        = ImageAddress(ir->image.band);
     in->image.pixname     = ImagePixName(ir->image.band);
     in->image.width       = ImageWidth(ir->image.band);
     in->image.height      = ImageHeight(ir->image.band);
     in->image.line_length = ImageLineLength(ir->image.band);
     in->image.xstart      = ImageXstart(ir->image.band);
     in->image.ystart      = ImageYstart(ir->image.band);
  }
  if (in->image.pixname)
    {
      pptr = iic->pixel_class;
      while (pptr && strcmp(pptr->pixelname, in->image.pixname)) 
	pptr = pptr->next;
    } else pptr = NULL;
  in->image.pixtype = pptr;
  in->image.update = TRUE;
  return(FALSE);
}




/*L*

________________________________________________________________

		SetValuesHook
________________________________________________________________

Name:		SetValuesHook
Syntax:		
Description:
Return value:
Author:		Otto Milvang
________________________________________________________________

*/


static Boolean SetValuesHook(wid, args, num_args)
Widget wid;
ArgList args;
Cardinal *num_args;
{
  ImageWidget iw = (ImageWidget) wid;
  int i, w, l;
  w = l = 0;
  FPRINTF(stderr, "Image: Set Values Hook: \n", 0,  0);
  for (i=0; i<*num_args; i++)
    {
      if (strcmp(args[i].name, XtNimageWidth)) w = i+1;
      if (strcmp(args[i].name, XtNimageLineLength)) l = 1;
    }
  if (w > 0 && l == 0 && (iw->image.pixtype != NULL)) 
    iw->image.line_length = 
      iw->image.width * (iw->image.pixtype)->pixelsize;
  Resize(iw);
  return(TRUE);
}



/*L*

________________________________________________________________

		Geometry
________________________________________________________________

Name:		Geometry
Syntax:		| static XtGeometryResult Geometry(wid, request, answer)
		|     Widget wid;
		|     XtWidgetGeometry *request, *answer;

Description:
Return value:
Author:		Otto Milvang
________________________________________________________________

*/

static XtGeometryResult Geometry(wid, request, answer)
Widget wid;
XtWidgetGeometry *request, *answer;
{
  FPRINTF(stderr, "Image: Geometry request flag %d \n", request->request_mode,0);
  if (request->request_mode & (CWWidth | CWHeight))
    {
      answer->width  = wid->core.width;
      answer->height = wid->core.height;
    }
  return(XtGeometryAlmost);
}





/*L*

________________________________________________________________

		ImageCoordinate
________________________________________________________________

Name:		ImageCoordinate
Syntax:		
Description:
Return value:
Author:		Otto Milvang
________________________________________________________________

*/



int ImageScreenXtoPos(wid, x)
ImageWidget wid;
int x;
{
  if (x<0 || x >= wid->core.width) return(-1);
  if (wid->image.resize)
    return(wid->image.xptr[x] + wid->image.xstart); else
    return(x + wid->image.xstart);
}

int ImageScreenYtoPos(wid, y)
ImageWidget wid;
int y;
{
  if (y<0 || y >= wid->core.height) return(-1);
  if (wid->image.resize)
    return(wid->image.yptr[y] + wid->image.ystart); else
    return(y + wid->image.ystart);
}

ImageData ImageScreenXYtoAdr(wid, x, y, options)
ImageWidget wid;
int x, y, options;
{
  int xa, ya;
  ImageData data;
  if (wid->image.pixtype == NULL) return(NULL);
  switch(options)
    {
    case 0: data = wid->image.data; break;
    case 1: data = wid->image.red_data; break;
    case 2: data = wid->image.green_data; break;
    case 3: data = wid->image.blue_data; break;
    }
  if (data == NULL) data = wid->image.data;
  if (x == -1)
    {
      xa = 0;
    } else {
      if (x < 0 || x >= wid->core.width)  return(NULL);
      xa = (wid->image.resize) ? (wid->image.xptr[x]) : x;
    }
  if (y < 0 || y >= wid->core.height) return(NULL);
  ya = (wid->image.resize) ? (wid->image.yptr[y]) : y;
  return((ImageData) (data + (wid->image.line_length) * ya
	+ xa * (wid->image.pixtype)->pixelsize));
}

ImageData ImagePosXYtoAdr(wid, x, y, options)
ImageWidget wid;
int x, y, options;
{
  ImageData data;
  if (wid->image.pixtype == NULL) return(NULL);
  switch(options)
    {
    case 0: data = wid->image.data; break;
    case 1: data = wid->image.red_data; break;
    case 2: data = wid->image.green_data; break;
    case 3: data = wid->image.blue_data; break;
    }
  if (data == NULL) data = wid->image.data;
  
  x -= wid->image.xstart;
  y -= wid->image.ystart;
  if (x < 0 || x >= wid->image.width) return(NULL);
  if (y < 0 || y >= wid->image.height) return(NULL);

  return((ImageData) (data + (wid->image.line_length) * y
	+ x * (wid->image.pixtype)->pixelsize));
}

int ImagePosXtoScreen(wid, x, right)
ImageWidget wid;
int x, right;
{
  int xx, *xarr, width;
  x -=  wid->image.xstart;
  if (wid->image.resize == 0) return(x);
  xarr = wid->image.xptr;
  width= wid->core.width;
  xx = (double) (x + (right ? 1 : 0) 
		 - wid->image.xpan) * wid->image.factor;
  if (xx < 0) return(xx);
  if (xx >= width) return(xx);
  if (right == 0)
    {
      while(xx>0 && xarr[xx-1] >= x) xx--; 
      while(xx<width && xarr[xx] < x) xx++; 
    } else {
      while(xx<width-1 && xarr[xx+1] <= x) xx++; 
      while(xx>=0 && xarr[xx] > x) xx--; 
    } 
  return(xx);
}

int ImagePosYtoScreen(wid, y, bottom)
ImageWidget wid;
int y, bottom;
{
  int yy, *yarr, width;
  y -= wid->image.ystart;
  if (wid->image.resize == 0) return(y);
  yarr = wid->image.yptr;
  width= wid->core.width;
  yy = (double) (y + (bottom ? 1 : 0) 
		 - wid->image.ypan) * wid->image.factor;
  if (yy < 0) return(yy);
  if (yy >= width) return(yy);
  if (bottom == 0)
    {
      while(yy>0 && yarr[yy-1] >= y) yy--; 
      while(yy<width && yarr[yy] < y) yy++; 
    } else {
      while(yy<width-1 && yarr[yy+1] <= y) yy++; 
      while(yy>=0 && yarr[yy] > y) yy--; 
    } 
  return(yy);
}




/*L*

________________________________________________________________

	        Button
________________________________________________________________

Name:		Button
Syntax:		
Description:
Return value:
Author:		Otto Milvang
________________________________________________________________

*/
static void Button(wid, event, params, num_params)
Widget wid;
XEvent *event;
String *params;
Cardinal *num_params;
{
  ImageWidget iw = (ImageWidget) wid;
  ImagePart *ip;  
  ImageCallbackRec button_callback;
  int x,y;
  ImageData data;

  ip = &(iw->image);
/*  fprintf(stderr, "Type=%6d,  Serial=%6d,  Time=%6d  ,State=%6d,\
	   Button=%6d, SS=%2d\n", event->xbutton.type,
	   event->xbutton.serial, event->xbutton.time,
	   event->xbutton.state, event->xbutton.button,
	   event->xbutton.same_screen); */
  data = NULL;
  if (event != NULL)
    {
      switch(event->type) {
      case ButtonPress:
      case ButtonRelease:
	x = event->xbutton.x;
	y = event->xbutton.y;
	break;
      case MotionNotify:
	x = event->xmotion.x;
	y = event->xmotion.y;
	break;
      default:
	x = -1;
	y = -1;
	break;
      }
    } else {
      x = -1;
      y = -1;
    }
  if (x >=0 && ip->pixtype)
    {
      data = (ImageData) ImageScreenXYtoAdr(iw, x, y, 0); 
      if (data == NULL) return;
    }
  button_callback.reason        = event ? ACTIVATE : RESIZE;
  button_callback.event         = event;
  button_callback.img           = ip->image;
  button_callback.band          = ip->band;
  button_callback.x             = ImageScreenXtoPos(iw, x);
  button_callback.y             = ImageScreenYtoPos(iw, y);
  button_callback.data          = data;
  button_callback.pixtype       = (pixel *) ip->pixtype;
  button_callback.zoom          = ip->zoom;
  button_callback.magnification = ip->factor;
  button_callback.xpan          = ip->xpan + ip->xstart;
  button_callback.ypan          = ip->ypan + ip->ystart;
  if (event)
    XtCallCallbacks(wid, XtNcallback, &button_callback); else  
      XtCallCallbacks(wid, XtNresizeCallback, &button_callback);  

} 




/*L*

________________________________________________________________

		Nop
________________________________________________________________

Name:		Nop
Syntax:		| Nop()
Description:	No operation
Return value:
Author:		Otto Milvang
________________________________________________________________

*/


static void Nop(wid, event, params, num_params)
ImageWidget wid;
XEvent *event;
String *params;
{}






/*L*

________________________________________________________________

		 ImageSetZoom
________________________________________________________________

Name:		ImageSetZoom
Syntax:         | void ImageSetZoom(wid, zoom, x, y)
		| Widget wid;
		| double zoom;
		| int x, y;
Description:
Return value:
Author:		Otto Milvang
________________________________________________________________

*/

void ImageTranslateZoomPanArgs(wid, zoomarg, panxarg, panyarg, mousex, mousey,
			  zoom, xpan, ypan)
ImageWidget wid;
char *zoomarg, *panxarg, *panyarg;
int  mousex, mousey,  *xpan, *ypan;
float *zoom;
{
  char *ptr, *val;
  int ww, hh;
  double atof();
  *zoom = wid->image.zoom;
  ptr = zoomarg;
  while (*ptr == ' ') ptr++;
  val = ptr;
  val++;
  switch(ptr[0])
    {
    case 'I': *zoom *= 2.0; break;
    case 'O': *zoom /= 2.0; break;
    case 'N': *zoom = 1.0;  break;
    case '+': *zoom += atof(val); break;
    case '-': *zoom -= atof(val); break;
    case '/': *zoom /= atof(val); break;
    case '*': *zoom *= atof(val); break;
    default:  *zoom = atof(ptr); break;
    }
  if (*zoom < 1.0) *zoom = 1.0;

  ptr = panxarg;
  while (*ptr == ' ') ptr++;
  val = ptr;
  val++;
  switch(ptr[0])
    {
    case 'M': *xpan = mousex;  break;
    case '+': *xpan += atoi(val); break;
    case '-': *xpan -= atoi(val); break;
    default:  *xpan = atoi(ptr); break;
    }
  if (ptr[strlen(ptr)-1] == 'C')
    {
      ww = (double) wid->image.width  / *zoom + 0.5;
      *xpan = *xpan - ww/2;
    }
  ptr = panyarg;
  while (*ptr == ' ') ptr++;
  val = ptr;
  val++;
  switch(ptr[0])
    {
    case 'M': *ypan = mousey;  break;
    case '+': *ypan += atoi(val); break;
    case '-': *ypan -= atoi(val); break;
    default:  *ypan = atoi(ptr); break;
    }
  if (ptr[strlen(ptr)-1] == 'C')
    {
      hh = (double) wid->image.height / *zoom + 0.5;
      *ypan = *ypan - hh/2;
    }
}


static void ZoomPan(wid, event, params, num_params)
ImageWidget wid;
XEvent *event;
String *params;
Cardinal *num_params;
{
  float zoom;
  int  xpan, ypan;
  ImageGetZoom(wid, &zoom, &xpan, &ypan, NULL);
  if (*num_params <= 3)
    {
      ImageTranslateZoomPanArgs(wid, params[0], params[1], params[2],
			  ImageScreenXtoPos(wid, event->xbutton.x),
			  ImageScreenYtoPos(wid, event->xbutton.y),
			  &zoom, &xpan, &ypan);
    } else ImageError(ERR_PARAM);
  FPRINTF(stderr, "Image: ZoomPan %.2f %d",zoom, xpan);
  FPRINTF(stderr, " %d\n",ypan, 0);
  ImageSetZoom(wid, zoom, xpan, ypan);
}



static void SetZoomParam(wid, zoom, x, y)
ImageWidget wid;
double zoom;
int x, y;
{
  int ww, hh, oldx, oldy;
  if (zoom <= 0.0) zoom = 1.0;
  if (zoom > wid->image.width ) zoom = wid->image.width;
  if (wid->image.factor*zoom/wid->image.zoom > 2047.99) return;
  FPRINTF(stderr, "Image: Zoom \n",zoom,0);
  
  ww = (double) wid->image.width  / zoom + 0.5;
  hh = (double) wid->image.height / zoom + 0.5;
  x = x - wid->image.xstart/* - ww/2 */;
  y = y - wid->image.ystart/* - hh/2 */; 
  if (x<0) x = 0;
  if (y<0) y = 0;
  if (x + ww > wid->image.width)  x = wid->image.width  - ww ; 
  if (y + hh > wid->image.height) y = wid->image.height - hh ; 
  if (wid->image.zoom == zoom && wid->image.xpan == x && 
      wid->image.ypan == y) return;
  
  wid->image.zoom = zoom;
  wid->image.xpan = x;
  wid->image.ypan = y;
}

void ImageSetZoom(wid, zoom, x, y)
ImageWidget wid;
double zoom;
int x, y;
{
  ImageWidget wid2;
  ImageClassPart *ic = &imageClassRec.image_class;
  int on, count, i;
  int w1, h1, w2, h2; 
  double factor;

  on = ic -> zoom_all;
  count = on ? ic->count : 1;
  SetZoomParam(wid, zoom, x, y);
  ImageRedisplay(wid, TRUE);
  factor = wid->image.factor;
  if (on == 0) return;
  
  w1 = (double) wid->core.width  / 2.0 / factor + 0.5;
  h1 = (double) wid->core.height / 2.0 / factor + 0.5;
  wid2 = wid;
  for (i=0; i<count; i++)  
    {
      wid2 = (ImageWidget) ic->widget_list[i];
      if (wid == wid2) continue;
      w2 = (double) wid2->core.width  / 2.0 / factor + 0.5;
      h2 = (double) wid2->core.height / 2.0 / factor + 0.5;
      SetZoomParam(wid2, zoom, x + w1 - w2, y + h1 - h2);
      ImageRedisplay(wid2, TRUE);
    }
}

void ImageGetZoom(wid, zoom, xpan, ypan, factor)
ImageWidget wid;
float *zoom;
int *xpan, *ypan;
double *factor;
{
  if (zoom) *zoom   = wid->image.zoom;
  if (xpan) *xpan = wid->image.xpan + wid->image.xstart;
  if (ypan) *ypan = wid->image.ypan + wid->image.ystart;
  if (factor) *factor = wid->image.factor;
}




/*L*

________________________________________________________________

		 ImageKill
________________________________________________________________

Name:		ImageKill
Syntax:		| void ImageKill(wid)
		| Widget wid
Description:
Return value:
Author:		Otto Milvang
________________________________________________________________

*/



void ImageKill(wid)
Widget wid;
{
  while (wid->core.parent) wid = wid->core.parent;
  XtDestroyWidget(wid);
}






/*L*

________________________________________________________________

		 Resize_factor
________________________________________________________________

Name:		Resize_factor
Syntax:		| static double Resize_factor(wid, ww, wh, xw, xh, bw, bh, update, zoom)
		| Widget wid;
		| int *ww, *wh, xw, xh, bw, bh, update, zoom;
Description:    | ww     - Requested Window Width  -> New Window Width
		| wh     - Requested Window Height -> New Window Height
		| xw     - Current Ximg width
		| xh     - Current Ximg height
		| bw     - Total band width
		| bh     - Total band height
		| update - Force to resize
		| zoom   - zoom factor
Return value:   | 0     - You don''t need to update
		| Other - Resize factor
Author:		Otto Milvang
________________________________________________________________

*/


static double Resize_new_ww_wh(wid, ww, wh, xw, xh, bw, bh, update, zoom)
Widget wid;
int *ww, *wh, xw, xh, bw, bh, update;
double zoom;
{
  double factor;

  /* If no resize, return */
  if (xh && xw)
    {
      if (*ww  == xw  && *wh == xh && update == 0)

      /* Resize  */
      if (1.0 * *ww / bw < 1.0 * *wh / bh)
	 {  
	   factor = 1.0 * *ww / bw;
	   *wh =  bh * factor;
	 } else {	
	   factor = 1.0 * *wh / bh;
	   *ww = bw * factor;
	 }

      /* Small changes in size, resize to original size */
      if ( Dist(*ww, xw) < 8 ) *ww = xw;
      if ( Dist(*wh, xh) < 8 ) *wh = xh;
      if (*ww  == xw  && *wh == xh && update == 0)
	 {
	   if (wid->core.width != *ww || wid->core.height != *wh)
	     {
		(void) XtMakeResizeRequest(wid, *ww, *wh, NULL, NULL);
		return(0.0);
	      }
	 }
    }

  /* Near Band width ? */
  if ( Dist(*ww, bw) <= 16 ) *ww = bw;
  if ( Dist(*wh, bh) <= 16 ) *wh = bh;


  /* Resize  */
  if (1.0 * *ww / bw < 1.0 * *wh / bh)
    {  
      factor = 1.0 * *ww / bw;
      *wh =  bh * factor;
    } else {	
      factor = 1.0 * *wh / bh;
      *ww = bw * factor;
    }

  /* New window size is *ww x *wh */
  return(factor*zoom);
}






/*L:ImageRedisplay*

________________________________________________________________

		ImageRedisplay
________________________________________________________________

Name:		ImageRedisplay
Syntax:		| void ImageRedisplay(wid, update)
                | ImageWidget wid;
                | int update;
Description:
Return value:
Author:		Otto Milvang
________________________________________________________________

*/


void ImageRedisplay(wid, update)
ImageWidget wid;
int update;
{
  WidgetClass wc;
  CoreClassPart *ic;
  FPRINTF(stderr, "Image: Image Redisplay %d..... \n",update,0);
  wc = XtClass(wid);
  ic = &wc->core_class;
  if (update) wid->image.update = update;
  (ic->resize)(wid);
  (ic->expose)(wid, NULL, NULL);
}




/*L:ImageRedisplayScreen*

________________________________________________________________

		ImageRedisplayScreen
________________________________________________________________

Name:		ImageRedisplayScreen
Syntax:		| void ImageRedisplayScreen(wid, x, y, width, height)
Description:
Return value:
Author:		Otto Milvang
________________________________________________________________

*/


void ImageRedisplayScreen(wid, x, y, width, height)
ImageWidget wid;
int x, y, width, height;
{
  WidgetClass wc;
  Region rect;
  CoreClassPart *ic;
  XPoint p[4];
  int update;
  ImagePart *ip;
  FPRINTF(stderr, "Image: Image Redisplay Screen %4d %4d ",x,y);
  FPRINTF(stderr, "%4d %4d\n",width, height);
  wc = XtClass(wid);
  ic = &wc->core_class;
  ip = &(wid->image);
  update = ip->update;
  ip->update = 2;
  ip->updatex = x;
  ip->updatey = y;
  ip->updatew = width;
  ip->updateh = height;
  (ic->resize)(wid);

  p[0].x = p[3].x = x;
  p[1].x = p[2].x = x + width;
  p[0].y = p[1].y = y;
  p[2].y = p[3].y = y + height;
  rect = XPolygonRegion(p, 4, EvenOddRule);
  (ic->expose)(wid, NULL, rect);
  ip->update = update;
}




/*L:ImageRedisplayArea*

________________________________________________________________

		ImageRedisplayArea
________________________________________________________________

Name:		ImageRedisplayArea
Syntax:		| int ImageRedisplayArea(wid, x, y, width, height)
Description:
Return value:
Author:		Otto Milvang
________________________________________________________________

*/


int ImageRedisplayArea(wid, x, y, width, height)
ImageWidget wid;
int x, y, width, height;
{
  int x1, x2, y1, y2, w, h;

  FPRINTF(stderr, "Image: Image Redisplay Area %d %d",x,y);
  FPRINTF(stderr, " %4d %4d\n", width, height);
  x1 = ImagePosXtoScreen(wid, x, 0);
  x2 = ImagePosXtoScreen(wid, x+width-1, 1);
  y1 = ImagePosYtoScreen(wid, y, 0);
  y2 = ImagePosYtoScreen(wid, y+height-1, 1);
  FPRINTF(stderr, "Image: Image Redisplay Area %d %d",x1,x2);
  FPRINTF(stderr, " %4d %4d\n", y1, y2);
  w = x2 - x1 + 1;
  h = y2 - y1 + 1;
  if (x1 >= wid->image.width || y1 >= wid->image.height) return(0);
  if (x1 < 0)
    {
      w += x1;
      x1 = 0;
    }
  if (y1 < 0)
    {
      h += y1;
      y1 = 0;
    }
  if (w < 0 || h < 0) return(0);
  if (x1 + w > wid->image.width) w = wid->image.width - x1;
  if (y1 + h > wid->image.width) h = wid->image.width - y1;
  ImageRedisplayScreen(wid, x1, y1, w, h);
  return(1);
}





/*L:ImageCopyImagePart*

________________________________________________________________

		ImageCopyImagePart
________________________________________________________________

Name:		ImageCopyImagePart
Syntax:		| void ImageCopyImagePart(widto, widfrom)
                | ImageWidget widto, widfrom;
Description:
Return value:
Author:		Otto Milvang
________________________________________________________________

*/




void ImageCopyImagePart(widto, widfrom)
ImageWidget widto, widfrom;
{
  widto->image = widfrom->image;
}




/*L:ImageSetZoomAll*

________________________________________________________________

		ImageSetZoomAll
________________________________________________________________

Name:		ImageSetZoomAll
Syntax:		| int ImageSetZoomAll(wid, on)
                | Widget wid;
                | int on;
Description:	Set zoom all flag, If 'on' = TRUE all widget
                will use the same zoom and pan parameters.
Return value:
Author:		Otto Milvang
________________________________________________________________

*/


int ImageSetZoomAll(wid, on)
Widget wid;
int on;
{
  ImageClassPart *ic = &imageClassRec.image_class;
  if (on >= 0) ic -> zoom_all = on;
  return(ic -> zoom_all);
}



/*L:ImageSetZoomAll*

________________________________________________________________

		ImageGetZoomAll
________________________________________________________________

Name:		ImageGetZoomAll
Syntax:		| ImageGetZoomAll(wid)
                | Widget wid;
Description:	Return zoom_all flag.
Return value:
Author:		Otto Milvang
________________________________________________________________

*/

ImageGetZoomAll(wid)
Widget wid;
{
  ImageClassPart *ic = &imageClassRec.image_class;
  return(ic -> zoom_all);
}
