 

/*

________________________________________________________________

        ImageOverlay.c
        %W% %E%, Copyright 1990, Blab, UiO
        Image processing lab, Department of Informatics
        University of Oslo
        E-mail: blab@uio.no
________________________________________________________________
  
  Permission to use, copy, modify and distribute this software and its
  documentation for any purpose and without fee is hereby granted, 
  provided that this copyright notice appear in all copies and that 
  both that copyright notice and this permission notice appear in supporting
  documentation and that the name of B-lab, Department of Informatics or
  University of Oslo not be used in advertising or publicity pertaining 
  to distribution of the software without specific, written prior permission.

  B-LAB DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL B-LAB
  BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
  OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
  CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 

*/

static char *SccsId = "%Z%%M% %I% %E%, Otto Milvang, Blab, UiO";



/*P:ImageOverlay*

________________________________________________________________

		ImageOverlay Widget Class
________________________________________________________________

Name:		ImageOverlay Widget Class
Description:	| Public Functions:
                |
                |     void ImageOverlayRedisplay(wid)
		|     Widget wid;
		| 
		|     void ImageOverlayKill(wid)
		|     Widget wid;
		|
		|     void Resize(wid)
		|     Widget wid;
                |
		|     void ImageOverlayZoom(wid, zoomin, x, y)
                |     Widget wid;
                |     int zoomin, x, y;
		|
		|     ImageData ImageOverlayScreenXYtoAdr(wid, x, y, option)
		|     ImageOverlayWidget wid;
		|     int x, y, option;
		|
		|     ImageData ImageOverlayPosXYtoAdr(wid, x, y, option)
		|     ImageOverlayWidget wid;
		|     int x, y, option;
		|
		|     int ImageOverlayTranslateX(wid,x)
		|     ImageOverlayWidget wid;
		|     int x;
		|
		|     int ImageOverlayTranslateY(wid,y)
		|     ImageOverlayWidget wid;
		|     int y;
		|
		|     ImageOverlayFillRectangle(wid, x, y, width,
		|           height, value, update)
		|     ImageOverlayWidget wid;
		|     ImageOverlayWidget wid;
		|     int x, y, width, height, value, update;
		|
		|     ImageOverlayDrawPoint(wid, x, y, size, value, update)
		|     ImageOverlayWidget wid;
		|     int x, y, size, value, update;
		|
		|     ImageOverlayDrawLine(wid, x1, y1, x2, y2, value, update)
		|     ImageOverlayWidget wid;
		|     int x1, y1, x2, y2, value, update;
		|
		|     ImageOverlayFill(wid, x, y, search,
		|          fill, nc, mode, update)
		|     ImageOverlayWidget wid;
		|     int x, y, search, fill, nc;
		|
		|
                | Private Functions:
		|
		|     static void ClassPartInitialize(wc)
		|     WidgetClass wc;
		|
                |     static void Initialize(request, new)
		|     Widget request, new;
		|
		|     static void Realize (wid, valueMask, attrs)
                |     Widget wid;
                |     XtValueMask *valueMask;
                |     XSetWindowAttributes *attrs;
		|
		|     static void Redisplay(wid, event, region)
		|     Widget wid;
		|     XEvent *event;
		|     Region region;
                |
		|     static void Resize(wid)
		|     Widget wid;
		|
		|     static XtGeometryResult Geometry(wid, request, answer)
		|     Widget wid;
		|     XtWidgetGeometry *request, *answer;
                |
		|     static Boolean SetValues(current, request, new)
		|     Widget current, request, new;
		|
		|     static void Kill(wid, event, params, num_params)
                |     Widget wid;
		|     XEvent *event;
		|     String *params;
		|     Cardinal *num_params;

Description:
Files:	
See also:	
Diagnostics:
Restrictions:
Return value:
Author:		Otto Milvang
Examples:
________________________________________________________________

*/





/*I*

________________________________________________________________

		Include files and definitions
________________________________________________________________

*/


#include <X11/Xos.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include "ImageOverlayP.h"
#include <stdio.h>
#include <ctype.h>


#ifdef DEBUG 
#define FPRINTF(a,b,c,d) fprintf(a,b,c,d)
#else
#define FPRINTF(a,b,c,d)
#endif




/*D*

________________________________________________________________

		Global data structure
________________________________________________________________

*/

char *malloc();


static void ClassPartInitialize();
static Boolean SetValues();
static Boolean SetValuesHook();
static void Initialize();
static void Realize();
static void Redisplay();
static XtGeometryResult Geometry();
static void Kill();
static void Resize();
int ImageOverlayResize();
char *ImageOverlayPrintf();

static pixel overlaypixel =
{
  "ImageOverlayPixel",
  1,
  ImageOverlayResize,
  ImageOverlayPrintf,
  3,
  SCALE_NONE,
 .0,
  0.0,
};


/* static XtActionsRec actions[] =
{
  / {name, procedure}, /
}; */


 static char translations[] =
"<Key>Q:             	kill()\n\
 None<Key>Left:		zoom_pan(+0, -4, +0)\n\
 None<Key>Right:	zoom_pan(+0, +4, +0)\n\
 None<Key>Up:		zoom_pan(+0, +0, -4)\n\
 None<Key>Down:		zoom_pan(+0, +0, +4)\n\
 Shift<Key>Left:	zoom_pan(+0, -16, +0)\n\
 Shift<Key>Right:	zoom_pan(+0, +16, +)\n\
 Shift<Key>Up:		zoom_pan(+0, +0, -16)\n\
 Shift<Key>Down:	zoom_pan(+0, +0, +16)\n\
 Ctrl<Key>Left:		zoom_pan(+0, -256, +0)\n\
 Ctrl<Key>Right:	zoom_pan(+0, +256, +0)\n\
 Ctrl<Key>Up:		zoom_pan(+0, +0, -256)\n\
 Ctrl<Key>Down:		zoom_pan(+0, +0, +256)\n\
 Shift Ctrl <Btn1Down>:	zoom_pan(*8, MousePosC, MousePosC)\n\
 Shift Ctrl <Btn2Down>: zoom_pan(32, MousePosC, MousePosC)\n\
 Shift Ctrl <Btn3Down>: zoom_pan(/8, MousePosC, MousePosC)\n\
 Ctrl <Btn1Down>:       zoom_pan(*2, MousePosC, MousePosC)\n\
 Ctrl <Btn2Down>:       zoom_pan(1, MousePosC, MousePosC)\n\
 Ctrl <Btn3Down>:       zoom_pan(/2, MousePosC, MousePosC)\n\
 None <Btn2Up>:      	notify()\n\
 None <Btn3Up>:      	notify()\n\
 <MouseMoved>:       	notify()";




static XtResource resources[] = {
#define ioffset(field) XtOffset(ImageOverlayWidget, image_overlay.field)
#define coffset(field) XtOffset(ImageOverlayWidget, core.field)
    /* {name, class, type, size, ioffset, default_type, default_addr}, */
    { XtNoverlay, XtCOverlay, XtRPointer, sizeof(char *),
	  ioffset(overlay), XtRPointer, NULL },
#undef ioffset
#undef coffset
};

#define SuperClass ((ImageWidgetClass)&imageClassRec)


ImageOverlayClassRec imageOverlayClassRec = {
  { /* core fields */
    /* superclass		*/	(WidgetClass) &imageClassRec,
    /* class_name		*/	"ImageOverlay",
    /* widget_size		*/	sizeof(ImageOverlayRec),
    /* class_initialize		*/	NULL,
    /* class_part_initialize	*/	ClassPartInitialize,
    /* class_inited		*/	FALSE,
    /* initialize		*/	Initialize,
    /* initialize_hook		*/	NULL,
    /* realize			*/	XtInheritRealize,
    /* actions			*/	/*actions*/ NULL,
    /* num_actions		*/	/* XtNumber(actions) */ 0,
    /* resources		*/	resources,
    /* num_resources		*/	XtNumber(resources),
    /* xrm_class		*/	NULLQUARK,
    /* compress_motion		*/	TRUE,
    /* compress_exposure	*/	TRUE,
    /* compress_enterleave	*/	TRUE,
    /* visible_interest		*/	FALSE,
    /* destroy			*/	NULL,
    /* resize			*/	Resize,
    /* expose			*/	Redisplay,
    /* set_values		*/	SetValues,
    /* set_values_hook		*/	NULL,
    /* set_values_almost	*/	XtInheritSetValuesAlmost,
    /* get_values_hook		*/	NULL,
    /* accept_focus		*/	NULL,
    /* version			*/	XtVersion,
    /* callback_private		*/	NULL,
    /* tm_table			*/	translations,
    /* query_geometry		*/	XtInheritQueryGeometry,
    /* display_accelerator	*/	XtInheritDisplayAccelerator,
    /* extension		*/	NULL
  },{
    /* composite_class fields */
    /* geometry_manager   	*/    	XtInheritGeometryManager,
    /* change_managed     	*/    	XtInheritChangeManaged,
    /* insert_child       	*/    	XtInheritInsertChild,
    /* delete_child       	*/    	XtInheritDeleteChild,
    /* extension          	*/    	NULL,
  },{
    /* image fields */
    /* count			*/	0,
    /* widget_list_length       */      0,
    /* widget_list              */      NULL,
    /* pixel class              */	NULL,
    /* zoom_all                 */      0,
  },{ 
    /* image overlay fields */
    /* dummy			*/	0,
  }
};

WidgetClass imageOverlayWidgetClass = (WidgetClass)&imageOverlayClassRec;





/*L*

________________________________________________________________

		SetValues
________________________________________________________________

Name:		SetValues
Syntax:		
Description:
Return value:
Author:		Otto Milvang
________________________________________________________________

*/

static Boolean SetValues(current, request, new)
Widget current, request, new;
{
  pixelptr pptr;
  ImageOverlayWidget ic = (ImageOverlayWidget) current;
  ImageOverlayWidget ir = (ImageOverlayWidget) request;
  ImageOverlayWidget in = (ImageOverlayWidget) new;
  
  int i, w, h;
  Display *dpy;
  int screen;

  dpy = XtDisplay(new);
  screen = DefaultScreen(dpy);


  FPRINTF(stderr, "Image: Set Overlay Values %10.2f %10.2f \n",in->image.minval,in->image.maxval);
  if (ic->image_overlay.overlay != ir->image_overlay.overlay)
    {
      in->image_overlay.data = ImageAddress(ir->image_overlay.overlay);
      in->image_overlay.line_length = 
	ImageLineLength(ir->image_overlay.overlay);
    }
  return(FALSE);
}



/*L*

________________________________________________________________

		Initialize
________________________________________________________________

Name:		Initialize
Syntax:		| static void Initialize(request, new)
                | Widget request, new;
		|
Description:
Return value:
Author:		Otto Milvang
________________________________________________________________

*/

static void ClassPartInitialize(wc)
WidgetClass wc;
{
  ImageInstallPixelType(&overlaypixel);
}

static void Initialize(request, new)
Widget request, new;
{
  ImageOverlayWidget iw = (ImageOverlayWidget) new;
  ImageOverlayPart *ip;
  ImageClassPart *ic = &imageClassRec.image_class;
  pixelptr pptr;
  int w, h;

  ip = &(iw->image_overlay);

  ip->data        = ip->overlay ? ImageAddress(ip->overlay) : NULL;
  ip->line_length = ip->overlay ? ImageLineLength(ip->overlay) : 
    iw->image.line_length;
  pptr = ic->pixel_class;
  while (pptr && strcmp(pptr->pixelname, "ImageOverlayPixel"))
    pptr = pptr->next;
  ip->pixtype = pptr;

  w = iw -> image.width;
  h = iw -> image.height;
  while (w > 1000 || h > 800) 
    {
      w >>= 1;
      h >>= 1;
    }
  if (strcmp((iw->core.widget_class)->core_class.class_name, "ImageOverlay") == 0)
    {
      if (iw->core.width == 0 || iw->core.height == 0)
	XtResizeWidget(new, w, h, 1); else
	Resize(new);
    }

/*  (*SuperClass->core_class.initialize)(request, new); */
  
  FPRINTF(stderr,"ImageOverlay: Initialize start\n",0,0);
}




/*L*

________________________________________________________________

		Redisplay 
________________________________________________________________

Name:		Redisplay, ImageOverlayRedisplay
Syntax:		| static void Redisplay(wid, event, region)
                | Widget wid;
		| XEvent *event;
		| Region region;
		|
		| void ImageOverlayRedisplay(wid, update)
		| Widget wid;
		| int update;
Description:    
Return value:
Author:		Otto Milvang
________________________________________________________________

*/


static void Redisplay(wid, event, region)
ImageOverlayWidget wid;
XEvent *event;
Region region;
{
  FPRINTF(stderr,"ImageOverlay: Redisplay\n",0,0);
  (*SuperClass->core_class.expose)(wid, event, region);
}  

void ImageOverlayRedisplay(wid, update)
ImageOverlayWidget wid;
int update;
{
  FPRINTF(stderr, "Image: Image Redisplay %d..... \n",update,0);
  if (update) wid->image.update = update;
  Resize(wid);
  Redisplay(wid, NULL, NULL);
}

void ImageOverlayCopyImagePart(widto, widfrom)
ImageOverlayWidget widto, widfrom;
{
  widto->image = widfrom->image;
}

void ImageOverlayCopyImageOverlayPart(widto, widfrom)
ImageOverlayWidget widto, widfrom;
{
  widto->image_overlay = widfrom->image_overlay;
}






/*L*

________________________________________________________________

		Resize
________________________________________________________________

Name:		Resize
Syntax:		| void Resize(wid)
                | Widget wid;
Description:
Return value:
Author:		Otto Milvang
________________________________________________________________

*/



static void Resize(wid)
ImageOverlayWidget wid;
{
  ImagePart *ip;
  int update;
  ip = &(wid->image);
  update = ip->update;
  FPRINTF(stderr,"ImageOverlay: Resize\n",0,0);
  (*SuperClass->core_class.resize)(wid);
  if (!wid->image_overlay.data) return;
  if (update != 2)
    {
      (wid->image_overlay.pixtype)->
	resize(wid, 0, 0, wid->core.width, wid->core.height);
    } else {
      (wid->image_overlay.pixtype)->
	resize(wid, ip->updatex, ip->updatey, ip->updatew, ip->updateh);
    }
}






/*F:ImageOverlayScreenXYtoAdr*

________________________________________________________________

		ImageOverlayScreenXYtoAdr
________________________________________________________________

Name:		ImageOverlayScreenXYtoAdr
Syntax:		| ImageData ImageOverlayScreenXYtoAdr(wid, x, y, option)
                | ImageOverlayWidget wid;
		| int x, y, option;
Description:
Return value:
Author:		Otto Milvang
________________________________________________________________

*/



ImageData ImageOverlayScreenXYtoAdr(wid, x, y, option)
ImageOverlayWidget wid;
int x, y, option;
{
  int xa, ya;
  ImageData data;
  data = (ImageData) wid->image_overlay.data;
  if (!data) return(NULL);
  if (x == -1)
    {
      xa = 0;
    } else {
      if (x < 0 || x >= wid->core.width)  return(NULL);
      xa = (wid->image.resize) ? (wid->image.xptr[x]) : x;
    }
  if (y < 0 || y >= wid->core.height) return(NULL);
  ya = (wid->image.resize) ? (wid->image.yptr[y]) : y;
  return((ImageData) (data + (wid->image_overlay.line_length) * ya
	+ xa * (wid->image_overlay.pixtype)->pixelsize));
}



/*F:ImageOverlayPosXYtoAdr*

________________________________________________________________

		ImageOverlayPosXYtoAdr
________________________________________________________________

Name:		ImageOverlayPosXYtoAdr
Syntax:		| ImageData ImageOverlayPosXYtoAdr(wid, x, y, option)
                | ImageOverlayWidget wid;
		| int x, y, option;
Description:
Return value:
Author:		Otto Milvang
________________________________________________________________

*/



ImageData ImageOverlayPosXYtoAdr(wid, x, y, option)
ImageOverlayWidget wid;
int x, y, option;
{
  ImageData data;
  data = (ImageData) wid->image_overlay.data;
  if (!data) return(NULL);
  x -= wid->image.xstart;
  y -= wid->image.ystart;

  if (x < 0 || x >= wid->image.width) return(NULL);
  if (y < 0 || y >= wid->image.height) return(NULL);
  return((ImageData) (data + (wid->image_overlay.line_length) * y
	+ x * (wid->image_overlay.pixtype)->pixelsize));
}






/*F:ImageOverlayFillRectangle*

________________________________________________________________

		ImageOverlayFillRectangle
________________________________________________________________

Name:		ImageOverlayFillRectangle
Syntax:		| ImageOverlayFillRectangle
                |   (wid, x, y, width, height, value, update)
                | ImageOverlayWidget wid;
                | int x, y, width, height, value, update;
Description:	Draw a filled rectangle in the overlay image at image 
                coordinates (x,y) with size (width, height).
		The rectangle  is filled with 'value' pixels. 
		The special value '-1' is used for drawing by inverting 
		the pixels. 'update' = TRUE will activate ImageRedisplayArea.
Return value:	
Author:		Otto Milvang
________________________________________________________________

*/


ImageOverlayFillRectangle(wid, x, y, width, height, value, update)
ImageOverlayWidget wid;
int x, y, width, height, value, update;
{
  ImageData data, pix;
  int xp, yp, xs, ys, ll, xmin, xmax, ymin, ymax;

  data = wid->image_overlay.data;
  if (!data) return;
  xs = wid->image.xstart;
  ys = wid->image.ystart;
  ll = wid->image_overlay.line_length;
  x -= xs;
  y -= ys;

  if (x < 0) 
    { 
      width -= (-x);
      x = 0;
    }
  if (y < 0) 
    { 
      height -= (-y);
      y = 0;
    }
  if (x+width  > wid->image.width)  width  = wid->image.width  - x;
  if (y+height > wid->image.height) height = wid->image.height - y;

  xmin = x + width;
  xmax = x;
  ymin = y + height;
  ymax = y;

  
  for (yp = y; yp < y+height; yp++)
    {
      pix = data + ll * yp + x;
      for (xp = x; xp < x+width; xp++, pix++)
	{
	  if (value == -1) *pix = ~(*pix); else
	    if (value != *pix)
	      {
		*pix = value;
		if (xp <xmin) xmin = xp;
		if (xp >xmax) xmax = xp;
		if (yp <ymin) ymin = yp;
		if (yp >ymax) ymax = yp;
	      }
	}
    }

  if (update)
    {
      if (value == -1)
	return(ImageRedisplayArea(wid, x + xs, y + ys, width, height)); else
	  if (xmin <= xmax || ymin <= ymax)
	    return(ImageRedisplayArea(
	      wid, xmin+xs, ymin+ys, xmax-xmin+1, ymax-ymin+1));
    } 
  return(0);
}





/*F:ImageOverlayDrawPoint*

________________________________________________________________

		ImageOverlayDrawPoint
________________________________________________________________

Name:		ImageOverlayDrawPoint
Syntax:		| ImageOverlayDrawPoint(wid, x, y, size, value, update)
                | ImageOverlayWidget wid;
                | int x, y, size, value, update;
Description:	Draw a point in the overlay image at image 
                coordinates (x,y). If size > 1 the point is
		written as a cross of size 'size'. The point 
		is filled with 'value' pixels. The special value
		'-1' is used for drawing by inverting the pixels.
		 'update' = TRUE will activate ImageRedisplayArea.
Return value:	
Author:		Otto Milvang
________________________________________________________________

*/


ImageOverlayDrawPoint(wid, x, y, size, value, update)
ImageOverlayWidget wid;
int x, y, size, value, update;
{
  ImageData pix, data;
  int u1, u2, i, xmin, xmax, ymin, ymax, xs, ys, ll;
  FPRINTF(stderr, "ImageDrawPoint (%4d %4d)\n", x, y);
  if (size > 0) size--; else size = 0;

  u1 = FALSE;
  data = wid->image_overlay.data;
  if (!data) return;
  xs = wid->image.xstart;
  ys = wid->image.ystart;
  ll = wid->image_overlay.line_length;
  x -= xs;
  y -= ys;
  xmin = x - size < 0 ? 0 : x - size;
  ymin = y - size < 0 ? 0 : y - size;
  xmax = x + size >= wid->image. width  ? wid->image.width-1  : x + size;
  ymax = x + size >= wid->image. height ? wid->image.height-1 : y + size;

  pix = data + ymin*ll +x;
  for (i=ymin; i<=ymax; i++, pix+=ll)
    if (value != *pix)
      {
	*pix = (value == -1) ? ~(*pix) : value;
	u1 = TRUE;
      }
  pix = data + y*ll +xmin;
  for (i=xmin; i<=xmax; i++, pix++)
    if (value != *pix)
      {
	*pix = (value == -1) ? ~(*pix) : value;
	u1 = TRUE;
      }
  if (update && u1)
    ImageRedisplayArea(wid, xmin+xs, ymin+ys, 2*size+1, 2*size+1);
  return(u1);
}	



/*F:ImageOverlayDrawLine*

________________________________________________________________

		ImageOverlayDrawLine
________________________________________________________________

Name:		ImageOverlayDrawLine
Syntax:		| ImageOverlayDrawLine(wid, x1, y1, x2, y2, value, update)
                | ImageOverlayWidget wid;
                | int x, y, x1, y1, x2, y2, value, update;
Description:	Draw a line in the overlay image from image 
                coordinates (x1,y1) to (x2,y2). The line 
		is written with 'value' pixels. The special value
		'-1' is used for drawing by inverting the pixels.
		'update' = TRUE will activate ImageRedisplayArea.
Return value:	
Author:		Otto Milvang
________________________________________________________________

*/


#define iabs(a) ((a) < 0 ? -(a) : (a))

ImageOverlayDrawLine(wid, x1, y1, x2, y2, value, update)
ImageOverlayWidget wid;
int x1, y1, x2, y2, value, update;
{
  double s1, f;
  int x, y, xl, yl, cl, w, xs, ys, ll, xmin, ymin, lstep, step, len, tl, u;
  ImageData pix, data;
  FPRINTF(stderr, "ImageDrawLine (%4d %4d)", x1, y1);
  FPRINTF(stderr, "-> (%4d %4d)\n", x2, y2);
  u = FALSE;
  data = wid->image_overlay.data;
  if (!data) return;
  xs = wid->image.xstart;
  ys = wid->image.ystart;
  ll = wid->image_overlay.line_length;
  x1 -= xs;
  x2 -= xs;
  y1 -= ys;
  y2 -= ys;
  xmin = x1 < x2 ? x1 : x2;
  ymin = y1 < y2 ? y1 : y2;

  xl = iabs(x2-x1)+1;
  yl = iabs(y2-y1)+1;
  if (xl <=2 && yl <=2)
    {
      for (pix = data + y1 * ll + x1, x=0; 
	   x<=1; x++, pix = data + y2 * ll + x2)
      if (value != *pix)
	{
	  *pix = value == -1 ? ~(*pix) : value;
	  u = TRUE;
	}
      if (update && u) ImageRedisplayArea(wid, xmin+xs, ymin+ys, xl, yl);
      return(u);
    }

  if (x1 == x2 || y1 == y2) 
    {
      if (x1 == x2) step = ll; else step = 1;
      len = xl > yl ? xl : yl;
      pix = data + ymin * ll + xmin;
      for (x=0; x<len; x++, pix+=step)
	if (value == -1) *pix = ~(*pix); else *pix = value;
      if (update) ImageRedisplayArea(wid, xmin+xs, ymin+ys, xl, yl);
      return(TRUE);
    }

  if (xl > yl) 
    {
      f = (double) xl / (double) yl;
      tl = xl;
      step = 1;
      lstep = (y2-y1)*(x2-x1) > 0 ? ll : -ll;
      pix = data + (xmin == x1 ? y1 : y2)*ll + xmin;
      len = yl;
      
    } else {
      f = (double) yl / (double) xl;
      tl = yl;
      step = ll;
      lstep = (y2-y1)*(x2-x1) > 0 ? 1 : -1;
      pix = data + ymin*ll + (ymin==y1 ? x1 : x2);
      len = xl;
    }
  cl = 0;
  s1 = 0.5;


  for (x = 0; x<len; x++, pix +=lstep)
    {
      s1 += f;
      w = (int) s1 - cl;
      if (x == len-1) w = tl-cl;
        for (y=0; y<w; y++, pix+=step)	
	  if (value == -1) *pix = ~(*pix); else *pix = value;
      cl += w;
    }
  if (update) ImageRedisplayArea(wid, xmin+xs, ymin+ys, xl, yl);
  return(TRUE);
}






/*F:ImageOverlayFill*

________________________________________________________________

		ImageOverlayFill
________________________________________________________________

Name:		ImageOverlayFill
Syntax:		| ImageOverlayFill(wid, x, y, search, fill, nc, mode, update)
                | ImageOverlayWidget wid;
		| int x, y, search, fill, nc;
Description:	Fill an area with 'fill' pixels. Start in position (x,y).
                Follow 'nc' nc = 4  (4 connected neighbours) or
		nc = 8 (8 connected neighbours).
		| 'mode' = 0, border fill. while (pix != search) fill; 
		| 'mode' = 1, fllod fill.  while (pix == search) fill;
		'update' = TRUE will activate ImageRedisplayArea.
Return value:
Author:		Otto Milvang
________________________________________________________________

*/


typedef struct
{
  ImageData data;
  int w, h, ll, xs, ys;
  int searchnr, fillnr, neighb, regionsize;
  int xmin, xmax, ymin, ymax, mode;
} filltype;

ImageOverlayFill(wid, x, y, search, fill, nc, mode, update)
ImageOverlayWidget wid;
int x, y, search, fill, nc;
{
  filltype info;
  FPRINTF(stderr, "ImageOverlayFill %3d %3d", x, y);
  FPRINTF(stderr, "  %3d %3d", search, fill);
  FPRINTF(stderr, "  %3d %3d\n", mode, update);
  info.data = wid->image_overlay.data;
  if (!info.data) return;
  info.xs = wid->image.xstart;
  info.ys = wid->image.ystart;
  info.ll = wid->image_overlay.line_length;
  info.w = wid->image.width;
  info.h = wid->image.height;
  x -= info.xs;
  y -= info.ys;

  info.searchnr = search;
  info.fillnr = fill;
  info.neighb = ( nc == 8 ) ? 1 : 0;        
  info.regionsize  = 0;
  info.mode = mode;
  info.xmin = x;
  info.xmax = x;
  info.ymin = y;
  info.ymax = y;


  switch(mode)
    {
    case 0: fillsegment(&info, x, y, x, x, -1); break;
    case 1: floodsegment(&info, x, y, x, x, -1); break;
    }
  if (info.regionsize && update)
    ImageRedisplayArea(wid, info.xmin + info.xs, info.ymin + info.ys,
		       info.xmax - info.xmin +1, info.ymax - info.ymin +1);
  return(info. regionsize);
}




#define REL !=
 		
static int fillsegment(info, x, y, xb, xe, down)			
filltype *info;
int x, y, xb, xe, down;					
{  							
    							
  int lx, x1, x2, ww, ll, searchnr, fillnr, r, y2, la, lb; 				
  ImageData data, ptr; 						
    			
  data = info->data;		
  ww = info->w-1;
  ll = info->ll;
  searchnr = info->searchnr;
  fillnr = info->fillnr;
  ptr = data + y*ll;
  
  if (!(ptr[x] REL searchnr)) return(0);
  for(x1=x;   x1 >=  0 && ptr[x1] REL searchnr; x1--) ptr[x1] = fillnr;
  for(x2=x+1; x2 <= ww && ptr[x2] REL searchnr; x2++) ptr[x2] = fillnr;
  x1 ++; x2--;
							
  info->regionsize += (x2-x1+1);	
  if ( x1 < info->xmin ) info->xmin = x1;
  if ( x2  > info->xmax ) info->xmax = x2;
  if ( y > info->ymax ) info->ymax = y;
  if ( y < info->ymin ) info->ymin = y;

  if ( info->neighb && (x1 > 0 ) ) x1--;     		
  if ( info->neighb && (x2 < ww) ) x2++;

  for (r = -1; r<=1; r++)
    {
      y2 = (r ==  1) ? y+down : y-down;
      la = (r ==  0) ? xe : x1;
      lb = (r == -1) ? xb : x2;  
      if ( y2 >= 0 &&  y2 < info->h)
	{
	  ptr = data + y2*ll;
	  for ( lx = la; lx <= lb; lx++ )		
	    if (ptr[lx]  REL  searchnr)		
	      lx += fillsegment(info, lx, y2, x1-1, x2+1, y2-y); 	
	}
    }
  return(x2-x);					
}						       

#undef REL



#define REL ==
 		
static int floodsegment(info, x, y, xb, xe, down)			
filltype *info;
int x, y, xb, xe, down;					
{  							
    							
  int lx, x1, x2, ww, ll, searchnr, fillnr, r, y2, la, lb; 				
  ImageData data, ptr; 						
    			
  data = info->data;		
  ww = info->w-1;
  ll = info->ll;
  searchnr = info->searchnr;
  fillnr = info->fillnr;
  ptr = data + y*ll;
  
  if (!(ptr[x] REL searchnr)) return(0);
  for(x1=x;   x1 >=  0 && ptr[x1] REL searchnr; x1--) ptr[x1] = fillnr;
  for(x2=x+1; x2 <= ww && ptr[x2] REL searchnr; x2++) ptr[x2] = fillnr;
  x1 ++; x2--;
							
  info->regionsize += (x2-x1+1);	
  if ( x1 < info->xmin ) info->xmin = x1;
  if ( x2  > info->xmax ) info->xmax = x2;
  if ( y > info->ymax ) info->ymax = y;
  if ( y < info->ymin ) info->ymin = y;

  if ( info->neighb && (x1 > 0 ) ) x1--;     		
  if ( info->neighb && (x2 < ww) ) x2++;

  for (r = -1; r<=1; r++)
    {
      y2 = (r ==  1) ? y+down : y-down;
      la = (r ==  0) ? xe : x1;
      lb = (r == -1) ? xb : x2;  
      if ( y2 >= 0 &&  y2 < info->h)
	{
	  ptr = data + y2*ll;
	  for ( lx = la; lx <= lb; lx++ )		
	    if (ptr[lx]  REL  searchnr)		
	      lx += floodsegment(info, lx, y2, x1-1, x2+1, y2-y); 	
	}
    }
  return(x2-x);					
}						       

#undef REL




