

/*

________________________________________________________________

        ImageProtoPix.c
        %W% %E%, Copyright 1990, Blab, UiO
        Image processing lab, Department of Informatics
        University of Oslo
        E-mail: blab@uio.no
________________________________________________________________
  
  Permission to use, copy, modify and distribute this software and its
  documentation for any purpose and without fee is hereby granted, 
  provided that this copyright notice appear in all copies and that 
  both that copyright notice and this permission notice appear in supporting
  documentation and that the name of B-lab, Department of Informatics or
  University of Oslo not be used in advertising or publicity pertaining 
  to distribution of the software without specific, written prior permission.

  B-LAB DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL B-LAB
  BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
  OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
  CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 

*/

static char *SccsId = "%Z%%M% %I% %E%, Blab, UiO";

#include <X11/Xos.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include "ImageP.h"
#include <stdio.h>
#include <ctype.h>

#ifdef DEBUG 
#define FPRINTF(a,b,c,d) fprintf(a,b,c,d)
#else
#define FPRINTF(a,b,c,d)
#endif

#define ERR_OK		0
#define ERR_MEM		1



/*F:ImageXxxxResize*

________________________________________________________________

		ImageXxxxResize
________________________________________________________________

Name:		ImageXxxxResize
Syntax:		| ImageXxxxResize(wid)
                | ImageWidget wid;
Description:	Resize the widget
Return value:
Author:		Otto Milvang
________________________________________________________________

*/

#ifndef R_NO
#define R_NO(a) *a++
#define R_YES(a) a[*xptr]
#define EQ_PRO(left, OP, right) left OP f1 * right + f2
#define EQ_24(a) EQ_PRO(*ptr++, =, a)
#define EQ_8(a) EQ_PRO(*ptr++, =, a)
#define EQ_4A(a) EQ_PRO(*ptr++, +=, a)
#define EQ_4B(a) EQ_PRO(j, =, a); *ptr = j << 4;
#define EQ_2A(a) EQ_PRO(0.0, <=, a)  
#define EQ_2B    = 
#endif


#ifndef INIT24
#define INIT24\
  f1 = scalea * 256.0/(maxval-minval);\
  f2 = (scaleb-minval) * 256.0/(maxval-minval) + 0.5;
#define INIT8 INIT24
#define INIT7\
  f1 = scalea * 128.0/(maxval-minval);\
  f2 = (scaleb-minval) * 128.0/(maxval-minval) + 64.5;
#define INIT4\
  f1 = scalea * 16.0/(maxval-minval);\
  f2 = (scaleb-minval) * 16.0/(maxval-minval) + 0.5;
#define INIT2\
  f1 = scalea /(maxval-minval);\
  f2 = (scaleb-minval) /(maxval-minval) - 0.5;
#endif

#ifndef ADDRESS
#define ADDRESS ImageScreenXYtoAdr
#endif

static int ResizeLSB();
static int ResizeMSB();
static int Resize4();
static int Resize7();
static int Resize8();
static int Resize24LSB();
static int Resize24MSB();

static  ImagePart *ip;
static  XImage *ximg;
static  int twidth, theight, lwidth, lheight, xstart, ystart, bpl, lpl; 
static  int resize, *xarr, *yarr;
static  ImagePixel *btr, *obtr;
static  unsigned char *ptr, *optr, *fptr;
static  int i, j, k, *xptr, x, y, xsize, pix, xfirst;
static  float scalea, scaleb, minval, maxval;

char *ImageXxxxPrintf(buf, pix)
char *buf;
ImagePixel *pix;
{
  sprintf(buf, PrintfFormat, *pix); 
  return(buf);
}


ImageXxxxResize(wid, x, y, width, height)
ImageWidget wid;
int x, y, width, height;
{

  xstart = x;
  ystart = y;
  lwidth = width;
  lheight = height; 
  ip = &(wid->image);
  ximg = ip->ximg;
  twidth = ximg->width;
  theight = ximg->height;
  resize = ip->resize;
  xfirst = resize ? -1 : xstart;
  xarr = ip->xptr;
  yarr = ip->yptr;
  scalea = ip->scalea;
  scaleb = ip->scaleb;
  minval = ip->minval;
  maxval = ip->maxval;
  optr = NULL;
  obtr = NULL;

  
  FPRINTF(stderr, "ImageProtoResize, %4d, %4d", xstart, ystart);
  FPRINTF(stderr, " %4d %4d\n", lwidth, lheight);
  switch(ip->displaymethod)
    {
    case TRUE_COLOR_24:
      if (ximg->byte_order == LSBFirst) Resize24LSB(wid);
      else Resize24MSB(wid);
      break;
    case STATIC_GRAY_8:
    case PSEUDO_COLOR_8_B:
      Resize8(wid);
      break;
    case GRAY_SCALE_8:
    case PSEUDO_COLOR_8_A:
      Resize7(wid);
      break;
    case STATIC_GRAY_4:
    case GRAY_SCALE_4:
      Resize4(wid);
      break;
    case STATIC_GRAY_1:
      if (ximg->bitmap_bit_order == LSBFirst) ResizeLSB(wid);
      else ResizeMSB(wid);
      break;
    default:
      Resize8(wid);
      break;

    }
}


static int Resize24LSB(wid)
ImageWidget wid;
{
  ImagePixel *r_btr, *g_btr, *b_btr;
  double f1, f2;
  FPRINTF(stderr, "Image: Resize 24 LSB, %d, %d\n", twidth, theight);
  bpl = ximg -> bytes_per_line = twidth*4;

  if (!ximg->data) ximg->data = (char *) malloc(bpl*theight);
  if (ximg->data == NULL) ImageError(wid, ERR_MEM);

  INIT24

  lpl = lwidth * 4;
  for(y = 0; y < lheight; y++)
    {
      r_btr =  (ImagePixel*) ADDRESS(wid, resize ? -1 : xstart, ystart+y, 1); 
      g_btr =  (ImagePixel*) ADDRESS(wid, resize ? -1 : xstart, ystart+y, 2); 
      b_btr =  (ImagePixel*) ADDRESS(wid, resize ? -1 : xstart, ystart+y, 3);
      if (obtr == r_btr) fptr = optr; else fptr = NULL; 
      obtr = r_btr;
      optr = ptr = (unsigned char *) &ximg->data[(ystart+y)*bpl+xstart*4]; 
      xptr = &xarr[xstart];
      xsize = lwidth;
      if (resize == 0)
	for(x = 0; x <lwidth; x++)
	  {	
	    EQ_24(R_NO(b_btr));
	    EQ_24(R_NO(g_btr));
	    EQ_24(R_NO(r_btr));
	    *ptr++ = 0;
	  } else {
	if (fptr != NULL) for (x=0; x<lpl; x++) *ptr++ = *fptr++; else
        for(x = 0; x <lwidth; x++, xptr++)
	  {
	    EQ_24(R_YES(b_btr));
	    EQ_24(R_YES(g_btr));
	    EQ_24(R_YES(r_btr));
	    *ptr++ = 0;
	  }
      }
    }
}


static int Resize24MSB(wid)
ImageWidget wid;
{
  ImagePixel *r_btr, *g_btr, *b_btr;
  double f1, f2;
  FPRINTF(stderr, "Image: Resize 24 MSB, %d, %d\n", twidth, theight);
  bpl = ximg -> bytes_per_line = twidth*4;

  if (!ximg->data) ximg->data = (char *) malloc(bpl*theight);
  if (ximg->data == NULL) ImageError(wid, ERR_MEM);

  INIT24

  lpl = lwidth * 4;
  for(y = 0; y < lheight; y++)
    {
      r_btr =  (ImagePixel*) ADDRESS(wid, resize ? -1 : xstart, ystart+y, 1); 
      g_btr =  (ImagePixel*) ADDRESS(wid, resize ? -1 : xstart, ystart+y, 2); 
      b_btr =  (ImagePixel*) ADDRESS(wid, resize ? -1 : xstart, ystart+y, 3);
      if (obtr == r_btr) fptr = optr; else fptr = NULL; 
      obtr = r_btr;
      optr = ptr = (unsigned char *) &ximg->data[(ystart+y)*bpl+xstart*4]; 
      xptr = &xarr[xstart];
      xsize = lwidth;
      if (resize == 0)
	for(x = 0; x <lwidth; x++)
	  {	
	    *ptr++ = 0;
	    EQ_24(R_NO(r_btr));
	    EQ_24(R_NO(g_btr));
	    EQ_24(R_NO(b_btr));
	  } else {
	if (fptr != NULL) for (x=0; x<lpl; x++) *ptr++ = *fptr++; else
        for(x = 0; x <lwidth; x++, xptr++)
	  {
	    *ptr++ = 0;
	    EQ_24(R_YES(r_btr));
	    EQ_24(R_YES(g_btr));
	    EQ_24(R_YES(b_btr));
	  }
      }
    }
}



static int Resize8(wid)
ImageWidget wid;
{
  double f1, f2;
  FPRINTF(stderr, "Image: Resize 8, %d, %d\n", twidth, theight);
  bpl = ximg -> bytes_per_line = twidth;

  if (!ximg->data) ximg->data = (char *) malloc(bpl*theight);
  if (ximg->data == NULL) ImageError(wid, ERR_MEM);
  ptr = (unsigned char *) ximg->data;

  INIT8

  lpl = lwidth;
  for(y = 0; y < lheight; y++)
    {
      btr =  (ImagePixel*) ADDRESS(wid, resize ? -1 : xstart, ystart+y, 0); 
      if (obtr == btr) fptr = optr; else fptr = NULL; 
      obtr = btr;
      optr = ptr = (unsigned char *) &ximg->data[(ystart+y)*bpl+xstart];
      xptr = &xarr[xstart];
      if (resize == 0)
	{ 
	  for(x = 0; x <lwidth; x++) EQ_8(R_NO(btr));
	} else { 
	if (fptr != NULL) for (x=0; x<lpl; x++) *ptr++ = *fptr++; else
	  for(x = 0; x <lwidth; x++, xptr++) EQ_8(R_YES(btr));
	}
    }
}




static int Resize7(wid)
ImageWidget wid;
{
  double f1, f2;
  FPRINTF(stderr, "Image: Resize 7, %d, %d\n", twidth, theight);
  FPRINTF(stderr, "Datastruct %08x\n", ximg->data, 0);
  bpl = ximg -> bytes_per_line = twidth;
  if (!ximg->data) ximg->data = (char *) malloc(bpl*theight);
  if (ximg->data == NULL) ImageError(wid, ERR_MEM);

  INIT7

  lpl = lwidth;
  for(y = 0; y < lheight; y++)
    {
      btr =  (ImagePixel*) ADDRESS(wid, resize ? -1 : xstart, ystart+y, 0);
      if (obtr == btr) fptr = optr; else fptr = NULL;
      obtr = btr;
      optr = ptr = (unsigned char *) &ximg->data[(ystart+y)*bpl+xstart];
      xptr = &xarr[xstart];
      if (resize == 0)
	{ 
	  for(x = 0; x <lwidth; x++) EQ_8(R_NO(btr)); 
	} else {
	  if (fptr != NULL) for (x=0; x<lpl; x++) *ptr++ = *fptr++; else
	  for(x = 0; x <lwidth; x++, xptr++) EQ_8(R_YES(btr));
	}
    }
}



static int Resize4(wid)
ImageWidget wid;
{
  double f1, f2;
  FPRINTF(stderr, "Image: Resize 4, %d, %d\n", twidth, theight);
  bpl = ximg -> bytes_per_line = (twidth+1)/2;
  if (!ximg->data) ximg->data = (char *) malloc(bpl*theight);
  if (ximg->data == NULL) ImageError(wid, ERR_MEM);
  ximg->byte_order = MSBFirst;

  INIT4

  if (xstart & 1) { xstart--; lwidth++; }
  if (lwidth & 1) if (xstart + lwidth < twidth) lwidth++; 

  lpl = (lwidth+1)/2;    
  for(y = 0; y < lheight; y++)
    {
      btr =  (ImagePixel*) ADDRESS(wid, resize ? -1 : xstart, ystart+y, 0); 
      if (obtr == btr) fptr = optr; else fptr = NULL; 
      obtr = btr;
      optr = ptr = (unsigned char *) &ximg->data[(ystart+y)* bpl +xstart/2];
      xptr = &xarr[xstart];
      if (resize == 0)
	{
	  for(x = 0; x <lwidth; x++, xptr++)
	    if (x & 1)
	      {
		EQ_4A(R_NO(btr));
	      } else {
		EQ_4B(R_NO(btr));
	      }
	} else {
	if (fptr != NULL) for (x=0; x<lpl; x++) *ptr++ = *fptr++; else
	  for(x = 0; x <lwidth; x++, xptr++)
	    if (x & 1)
	      {	
		EQ_4A(R_YES(btr));
	      } else {
		EQ_4B(R_YES(btr));
	      }
	}
      if (lwidth & 1) ptr++;
    }
}	




static int ResizeMSB(wid)
ImageWidget wid;
{
  int count, w;
  double f1, f2;
  FPRINTF(stderr, "Image: Resize MSB, %d, %d\n", twidth, theight);
  bpl = ximg -> bytes_per_line = (twidth+7)/8;
  if (!ximg->data) ximg->data = (char *) malloc(bpl*theight);
  if (ximg->data == NULL) ImageError(wid, ERR_MEM);

  INIT2

  count = xstart & 7;
  if (count) 
    {
      xstart -= count;
      lwidth += count;
    }    
  if (lwidth & 7) lwidth = lwidth - (lwidth & 7) + 8;
  if (xstart + lwidth > twidth) lwidth = twidth-xstart; 
  FPRINTF(stderr, "Image: Resize MSB, %d, %d\n", lwidth, xstart);
  lpl = (lwidth+7)/8;
  for(y = 0; y < lheight; y++)
    {
      btr =  (ImagePixel*) ADDRESS(wid, resize ? -1 : xstart, ystart+y, 0); 
      if (obtr == btr) fptr = optr; else fptr = NULL; 
      obtr = btr;
      optr = ptr = (unsigned char *) &ximg->data[(ystart+y)* bpl +xstart/8];
      xptr = &xarr[xstart];
      count = 8;
      pix = 0;
      if (resize == 0)
	for(x = 0; x <lwidth; x++)
	  {
	    pix <<= 1;
	    if (EQ_2A(R_NO(btr))) pix++; 
	    if (--count == 0) 
	      {
		*ptr++ EQ_2B pix;
		count = 8;
		pix = 0;
	      }
	  } else 
	if (fptr != NULL) for (x=0; x<lpl; x++) *ptr++ = *fptr++; else
	for(x = 0; x <lwidth; x++, xptr++)
	  {
	    pix <<= 1;
	    if (EQ_2A(R_YES(btr))) pix++;
	    if (--count == 0) 
	      {
		*ptr++ EQ_2B pix;
		count = 8;
		pix = 0;
	      }
	  }
      if (count < 8)
	{
	  pix <<= count;
	  *ptr++ EQ_2B pix;
	}
    }
}




static int ResizeLSB(wid)
ImageWidget wid;
{
  int count, w;
  double f1, f2;

  FPRINTF(stderr, "Image: Resize LSB, %d, %d\n", twidth, theight);
  bpl = ximg -> bytes_per_line = (twidth+7)/8;
  if (!ximg->data) ximg->data = (char *) malloc(bpl*theight);
  if (ximg->data == NULL) ImageError(wid, ERR_MEM);

  INIT2

  count = xstart & 7;
  if (count) 
    {
      xstart -= count;
      lwidth += count;
    }    
  if (lwidth & 7) lwidth = lwidth - (lwidth & 7) + 8;
  if (xstart + lwidth > twidth) lwidth = twidth-xstart; 
  FPRINTF(stderr, "Image: Resize LSB, %d, %d\n", lwidth, xstart);
  lpl = (lwidth+7)/8;
  for(y = 0; y < lheight; y++)
    {
      btr =  (ImagePixel*) ADDRESS(wid, resize ? -1 : xstart, ystart+y, 0);
      if (obtr == btr) fptr = optr; else fptr = NULL; 
      obtr = btr;
      optr = ptr = (unsigned char *) &ximg->data[(ystart+y)* bpl +xstart/8];
      xptr = &xarr[xstart];
      count = 8;
      pix = 0;
      if (resize == 0)
	for(x = 0; x <lwidth; x++)
	  {
	    pix >>= 1;
	    if (EQ_2A(R_NO(btr))) pix |= 0x80;
	    if (--count == 0) 
	      {
		*ptr++ EQ_2B pix;
		count = 8;
		pix = 0;
	      }
	  } else 
	if (fptr != NULL) for (x=0; x<lpl; x++) *ptr++ = *fptr++; else
	for(x = 0; x <lwidth; x++, xptr++)
	  {
	    pix >>= 1;
	    if (EQ_2A(R_YES(btr))) pix |= 0x80;
	    if (--count == 0) 
	      {
		*ptr++ EQ_2B pix;
		count = 8;
		pix = 0;
	      }
	  }
      if (count < 8)
	{
	  pix <<= count;
	  *ptr++ EQ_2B pix;
	}
    }
}

